@@ -3,7 +3,7 @@
 const inherits = require('util').inherits;
 const EventEmitter = require('events').EventEmitter;
 const MongoError = require('../error').MongoError;
-const MongoNetworkError = require('../error').MongoNetworkError;
+const MongoTimeoutError = require('../error').MongoTimeoutError;
 const MongoWriteConcernError = require('../error').MongoWriteConcernError;
 const Logger = require('./logger');
 const f = require('util').format;
@@ -20,12 +20,22 @@ const Buffer = require('safe-buffer').Buffer;
 const connect = require('./connect');
 const updateSessionFromResponse = require('../sessions').updateSessionFromResponse;
 const eachAsync = require('../utils').eachAsync;
+const makeStateMachine = require('../utils').makeStateMachine;
 
-var DISCONNECTED = 'disconnected';
-var CONNECTING = 'connecting';
-var CONNECTED = 'connected';
-var DESTROYING = 'destroying';
-var DESTROYED = 'destroyed';
+const DISCONNECTED = 'disconnected';
+const CONNECTING = 'connecting';
+const CONNECTED = 'connected';
+const DRAINING = 'draining';
+const DESTROYING = 'destroying';
+const DESTROYED = 'destroyed';
+const stateTransition = makeStateMachine({
+  [DISCONNECTED]: [CONNECTING, DRAINING, DISCONNECTED],
+  [CONNECTING]: [CONNECTING, CONNECTED, DRAINING, DISCONNECTED],
+  [CONNECTED]: [CONNECTED, DISCONNECTED, DRAINING],
+  [DRAINING]: [DRAINING, DESTROYING, DESTROYED],
+  [DESTROYING]: [DESTROYING, DESTROYED],
+  [DESTROYED]: [DESTROYED]
+});
 
 const CONNECTION_EVENTS = new Set([
   'error',
@@ -60,7 +70,7 @@ var _id = 0;
  * @param {Buffer} [options.crl] SSL Certificate revocation store binary buffer
  * @param {Buffer} [options.cert] SSL Certificate binary buffer
  * @param {Buffer} [options.key] SSL Key file binary buffer
- * @param {string} [options.passPhrase] SSL Certificate pass phrase
+ * @param {string} [options.passphrase] SSL Certificate pass phrase
  * @param {boolean} [options.rejectUnauthorized=false] Reject unauthorized server certificates
  * @param {boolean} [options.promoteLongs=true] Convert Long values from the db into Numbers if they fit into 53 bits
  * @param {boolean} [options.promoteValues=true] Promotes BSON values to native types where possible, set to false to only receive wrapper types.
@@ -80,6 +90,14 @@ var Pool = function(topology, options) {
   // Store topology for later use
   this.topology = topology;
 
+  this.s = {
+    state: DISCONNECTED,
+    cancellationToken: new EventEmitter()
+  };
+
+  // we don't care how many connections are listening for cancellation
+  this.s.cancellationToken.setMaxListeners(Infinity);
+
   // Add the options
   this.options = Object.assign(
     {
@@ -103,7 +121,7 @@ var Pool = function(topology, options) {
       crl: null,
       cert: null,
       key: null,
-      passPhrase: null,
+      passphrase: null,
       rejectUnauthorized: false,
       promoteLongs: true,
       promoteValues: true,
@@ -113,7 +131,9 @@ var Pool = function(topology, options) {
       reconnectInterval: 1000,
       reconnectTries: 30,
       // Enable domains
-      domainsEnabled: false
+      domainsEnabled: false,
+      // feature flag for determining if we are running with the unified topology or not
+      legacyCompatMode: true
     },
     options
   );
@@ -123,6 +143,7 @@ var Pool = function(topology, options) {
   // Current reconnect retries
   this.retriesLeft = this.options.reconnectTries;
   this.reconnectId = null;
+  this.reconnectError = null;
   // No bson parser passed in
   if (
     !options.bson ||
@@ -135,8 +156,6 @@ var Pool = function(topology, options) {
 
   // Logger instance
   this.logger = Logger('Pool', options);
-  // Pool state
-  this.state = DISCONNECTED;
   // Connections
   this.availableConnections = [];
   this.inUseConnections = [];
@@ -146,9 +165,6 @@ var Pool = function(topology, options) {
   // Operation work queue
   this.queue = [];
 
-  // Contains the reconnect connection
-  this.reconnectConnection = null;
-
   // Number of consecutive timeouts caught
   this.numberOfConsecutiveTimeouts = 0;
   // Current pool Index
@@ -208,60 +224,39 @@ Object.defineProperty(Pool.prototype, 'socketTimeout', {
   }
 });
 
+Object.defineProperty(Pool.prototype, 'state', {
+  enumerable: true,
+  get: function() {
+    return this.s.state;
+  }
+});
+
 // clears all pool state
 function resetPoolState(pool) {
   pool.inUseConnections = [];
   pool.availableConnections = [];
   pool.connectingConnections = 0;
   pool.executing = false;
-  pool.reconnectConnection = null;
   pool.numberOfConsecutiveTimeouts = 0;
   pool.connectionIndex = 0;
   pool.retriesLeft = pool.options.reconnectTries;
   pool.reconnectId = null;
 }
 
-function stateTransition(self, newState) {
-  var legalTransitions = {
-    disconnected: [CONNECTING, DESTROYING, DISCONNECTED],
-    connecting: [CONNECTING, DESTROYING, CONNECTED, DISCONNECTED],
-    connected: [CONNECTED, DISCONNECTED, DESTROYING],
-    destroying: [DESTROYING, DESTROYED],
-    destroyed: [DESTROYED]
-  };
-
-  // Get current state
-  var legalStates = legalTransitions[self.state];
-  if (legalStates && legalStates.indexOf(newState) !== -1) {
-    self.emit('stateChanged', self.state, newState);
-    self.state = newState;
-  } else {
-    self.logger.error(
-      f(
-        'Pool with id [%s] failed attempted illegal state transition from [%s] to [%s] only following state allowed [%s]',
-        self.id,
-        self.state,
-        newState,
-        legalStates
-      )
-    );
-  }
-}
-
 function connectionFailureHandler(pool, event, err, conn) {
   if (conn) {
-    if (conn._connectionFailHandled) return;
+    if (conn._connectionFailHandled) {
+      return;
+    }
+
     conn._connectionFailHandled = true;
     conn.destroy();
 
     // Remove the connection
     removeConnection(pool, conn);
 
-    // Flush all work Items on this connection
-    while (conn.workItems.length > 0) {
-      const workItem = conn.workItems.shift();
-      if (workItem.cb) workItem.cb(err);
-    }
+    // flush remaining work items
+    conn.flush(err);
   }
 
   // Did we catch a timeout, increment the numberOfConsecutiveTimeouts
@@ -281,8 +276,10 @@ function connectionFailureHandler(pool, event, err, conn) {
 
   // No more socket available propegate the event
   if (pool.socketCount() === 0) {
-    if (pool.state !== DESTROYED && pool.state !== DESTROYING) {
-      stateTransition(pool, DISCONNECTED);
+    if (pool.state !== DESTROYED && pool.state !== DESTROYING && pool.state !== DRAINING) {
+      if (pool.options.reconnect) {
+        stateTransition(pool, DISCONNECTED);
+      }
     }
 
     // Do not emit error events, they are always close events
@@ -293,73 +290,60 @@ function connectionFailureHandler(pool, event, err, conn) {
 
   // Start reconnection attempts
   if (!pool.reconnectId && pool.options.reconnect) {
+    pool.reconnectError = err;
     pool.reconnectId = setTimeout(attemptReconnect(pool), pool.options.reconnectInterval);
   }
 
   // Do we need to do anything to maintain the minimum pool size
   const totalConnections = totalConnectionCount(pool);
   if (totalConnections < pool.minSize) {
-    _createConnection(pool);
+    createConnection(pool);
   }
 }
 
-function attemptReconnect(self) {
+function attemptReconnect(pool, callback) {
   return function() {
-    self.emit('attemptReconnect', self);
-    if (self.state === DESTROYED || self.state === DESTROYING) return;
+    pool.emit('attemptReconnect', pool);
+
+    if (pool.state === DESTROYED || pool.state === DESTROYING) {
+      if (typeof callback === 'function') {
+        callback(new MongoError('Cannot create connection when pool is destroyed'));
+      }
 
-    // We are connected do not try again
-    if (self.isConnected()) {
-      self.reconnectId = null;
       return;
     }
 
-    self.connectingConnections++;
-    connect(self.options, (err, connection) => {
-      self.connectingConnections--;
+    pool.retriesLeft = pool.retriesLeft - 1;
+    if (pool.retriesLeft <= 0) {
+      pool.destroy();
 
-      if (err) {
-        if (self.logger.isDebug()) {
-          self.logger.debug(`connection attempt failed with error [${JSON.stringify(err)}]`);
-        }
+      const error = new MongoTimeoutError(
+        `failed to reconnect after ${pool.options.reconnectTries} attempts with interval ${pool.options.reconnectInterval} ms`,
+        pool.reconnectError
+      );
 
-        self.retriesLeft = self.retriesLeft - 1;
-        if (self.retriesLeft <= 0) {
-          self.destroy();
-          self.emit(
-            'reconnectFailed',
-            new MongoNetworkError(
-              f(
-                'failed to reconnect after %s attempts with interval %s ms',
-                self.options.reconnectTries,
-                self.options.reconnectInterval
-              )
-            )
-          );
-        } else {
-          self.reconnectId = setTimeout(attemptReconnect(self), self.options.reconnectInterval);
-        }
-
-        return;
+      pool.emit('reconnectFailed', error);
+      if (typeof callback === 'function') {
+        callback(error);
       }
 
-      if (self.state === DESTROYED || self.state === DESTROYING) {
-        return connection.destroy();
+      return;
+    }
+
+    // clear the reconnect id on retry
+    pool.reconnectId = null;
+
+    // now retry creating a connection
+    createConnection(pool, (err, conn) => {
+      if (err == null) {
+        pool.reconnectId = null;
+        pool.retriesLeft = pool.options.reconnectTries;
+        pool.emit('reconnect', pool);
       }
 
-      self.reconnectId = null;
-      handlers.forEach(event => connection.removeAllListeners(event));
-      connection.on('error', self._connectionErrorHandler);
-      connection.on('close', self._connectionCloseHandler);
-      connection.on('timeout', self._connectionTimeoutHandler);
-      connection.on('parseError', self._connectionParseErrorHandler);
-      connection.on('message', self._messageHandler);
-
-      self.retriesLeft = self.options.reconnectTries;
-      self.availableConnections.push(connection);
-      self.reconnectConnection = null;
-      self.emit('reconnect', self);
-      _execute(self)();
+      if (typeof callback === 'function') {
+        callback(err, conn);
+      }
     });
   };
 }
@@ -448,7 +432,7 @@ function messageHandler(self) {
           updateSessionFromResponse(session, document);
         }
 
-        if (document.$clusterTime) {
+        if (self.topology && document.$clusterTime) {
           self.topology.clusterTime = document.$clusterTime;
         }
       }
@@ -559,69 +543,42 @@ Pool.prototype.isDisconnected = function() {
 /**
  * Connect pool
  */
-Pool.prototype.connect = function() {
+Pool.prototype.connect = function(callback) {
   if (this.state !== DISCONNECTED) {
     throw new MongoError('connection in unlawful state ' + this.state);
   }
 
-  const self = this;
   stateTransition(this, CONNECTING);
-
-  self.connectingConnections++;
-  connect(self.options, (err, connection) => {
-    self.connectingConnections--;
-
+  createConnection(this, (err, conn) => {
     if (err) {
-      if (self.logger.isDebug()) {
-        self.logger.debug(`connection attempt failed with error [${JSON.stringify(err)}]`);
+      if (typeof callback === 'function') {
+        this.destroy();
+        callback(err);
+        return;
       }
 
-      if (self.state === CONNECTING) {
-        self.emit('error', err);
+      if (this.state === CONNECTING) {
+        this.emit('error', err);
       }
 
+      this.destroy();
       return;
     }
 
-    if (self.state === DESTROYED || self.state === DESTROYING) {
-      return self.destroy();
-    }
+    stateTransition(this, CONNECTED);
 
-    // attach event handlers
-    connection.on('error', self._connectionErrorHandler);
-    connection.on('close', self._connectionCloseHandler);
-    connection.on('timeout', self._connectionTimeoutHandler);
-    connection.on('parseError', self._connectionParseErrorHandler);
-    connection.on('message', self._messageHandler);
-
-    // If we are in a topology, delegate the auth to it
-    // This is to avoid issues where we would auth against an
-    // arbiter
-    if (self.options.inTopology) {
-      stateTransition(self, CONNECTED);
-      self.availableConnections.push(connection);
-      return self.emit('connect', self, connection);
-    }
-
-    if (self.state === DESTROYED || self.state === DESTROYING) {
-      return self.destroy();
-    }
-
-    if (err) {
-      self.destroy();
-      return self.emit('error', err);
-    }
-
-    stateTransition(self, CONNECTED);
-    self.availableConnections.push(connection);
-
-    if (self.minSize) {
-      for (let i = 0; i < self.minSize; i++) {
-        _createConnection(self);
+    // create min connections
+    if (this.minSize) {
+      for (let i = 0; i < this.minSize; i++) {
+        createConnection(this);
       }
     }
 
-    self.emit('connect', self, connection);
+    if (typeof callback === 'function') {
+      callback(null, conn);
+    } else {
+      this.emit('connect', this, conn);
+    }
   });
 };
 
@@ -656,6 +613,11 @@ Pool.prototype.unref = function() {
 
 // Destroy the connections
 function destroy(self, connections, options, callback) {
+  stateTransition(self, DESTROYING);
+
+  // indicate that in-flight connections should cancel
+  self.s.cancellationToken.emit('cancel');
+
   eachAsync(
     connections,
     (conn, cb) => {
@@ -663,6 +625,9 @@ function destroy(self, connections, options, callback) {
         conn.removeAllListeners(eventName);
       }
 
+      // ignore any errors during destruction
+      conn.on('error', () => {});
+
       conn.destroy(options, cb);
     },
     err => {
@@ -686,14 +651,19 @@ function destroy(self, connections, options, callback) {
  */
 Pool.prototype.destroy = function(force, callback) {
   var self = this;
+  if (typeof force === 'function') {
+    callback = force;
+    force = false;
+  }
+
   // Do not try again if the pool is already dead
   if (this.state === DESTROYED || self.state === DESTROYING) {
     if (typeof callback === 'function') callback(null, null);
     return;
   }
 
-  // Set state to destroyed
-  stateTransition(this, DESTROYING);
+  // Set state to draining
+  stateTransition(this, DRAINING);
 
   // Are we force closing
   if (force) {
@@ -718,14 +688,16 @@ Pool.prototype.destroy = function(force, callback) {
     clearTimeout(this.reconnectId);
   }
 
-  // If we have a reconnect connection running, close
-  // immediately
-  if (this.reconnectConnection) {
-    this.reconnectConnection.destroy();
-  }
-
   // Wait for the operations to drain before we close the pool
   function checkStatus() {
+    if (self.state === DESTROYED || self.state === DESTROYING) {
+      if (typeof callback === 'function') {
+        callback();
+      }
+
+      return;
+    }
+
     flushMonitoringOperations(self.queue);
 
     if (self.queue.length === 0) {
@@ -742,7 +714,6 @@ Pool.prototype.destroy = function(force, callback) {
       }
 
       destroy(self, connections, { force: false }, callback);
-      // } else if (self.queue.length > 0 && !this.reconnectId) {
     } else {
       // Ensure we empty the queue
       _execute(self)();
@@ -761,6 +732,18 @@ Pool.prototype.destroy = function(force, callback) {
  * @param {function} [callback]
  */
 Pool.prototype.reset = function(callback) {
+  if (this.s.state !== CONNECTED) {
+    if (typeof callback === 'function') {
+      callback(new MongoError('pool is not connected, reset aborted'));
+    }
+
+    return;
+  }
+
+  // signal in-flight connections should be cancelled
+  this.s.cancellationToken.emit('cancel');
+
+  // destroy existing connections
   const connections = this.availableConnections.concat(this.inUseConnections);
   eachAsync(
     connections,
@@ -781,12 +764,12 @@ Pool.prototype.reset = function(callback) {
 
       resetPoolState(this);
 
-      // create an initial connection, and kick off execution again
-      _createConnection(this);
-
-      if (typeof callback === 'function') {
-        callback(null, null);
-      }
+      // create a new connection, this will ultimately trigger execution
+      createConnection(this, () => {
+        if (typeof callback === 'function') {
+          callback(null, null);
+        }
+      });
     }
   );
 };
@@ -854,17 +837,12 @@ Pool.prototype.write = function(command, options, cb) {
 
   // Pool was destroyed error out
   if (this.state === DESTROYED || this.state === DESTROYING) {
-    // Callback with an error
-    if (cb) {
-      try {
-        cb(new MongoError('pool destroyed'));
-      } catch (err) {
-        process.nextTick(function() {
-          throw err;
-        });
-      }
-    }
+    cb(new MongoError('pool destroyed'));
+    return;
+  }
 
+  if (this.state === DRAINING) {
+    cb(new MongoError('pool is draining, new operations prohibited'));
     return;
   }
 
@@ -912,10 +890,6 @@ Pool.prototype.write = function(command, options, cb) {
   // Optional per operation socketTimeout
   operation.socketTimeout = options.socketTimeout;
   operation.monitoring = options.monitoring;
-  // Custom socket Timeout
-  if (options.socketTimeout) {
-    operation.socketTimeout = options.socketTimeout;
-  }
 
   // Get the requestId
   operation.requestId = command.requestId;
@@ -978,7 +952,7 @@ Pool.prototype.write = function(command, options, cb) {
 function canCompress(command) {
   const commandDoc = command instanceof Msg ? command.command : command.query;
   const commandName = Object.keys(commandDoc)[0];
-  return uncompressibleCommands.indexOf(commandName) === -1;
+  return !uncompressibleCommands.has(commandName);
 }
 
 // Remove connection method
@@ -996,55 +970,73 @@ function removeConnection(self, connection) {
   if (remove(connection, self.inUseConnections)) return;
 }
 
-const handlers = ['close', 'message', 'error', 'timeout', 'parseError', 'connect'];
-function _createConnection(self) {
-  if (self.state === DESTROYED || self.state === DESTROYING) {
+function createConnection(pool, callback) {
+  if (pool.state === DESTROYED || pool.state === DESTROYING) {
+    if (typeof callback === 'function') {
+      callback(new MongoError('Cannot create connection when pool is destroyed'));
+    }
+
     return;
   }
 
-  self.connectingConnections++;
-  connect(self.options, (err, connection) => {
-    self.connectingConnections--;
+  pool.connectingConnections++;
+  connect(pool.options, pool.s.cancellationToken, (err, connection) => {
+    pool.connectingConnections--;
 
     if (err) {
-      if (self.logger.isDebug()) {
-        self.logger.debug(`connection attempt failed with error [${JSON.stringify(err)}]`);
+      if (pool.logger.isDebug()) {
+        pool.logger.debug(`connection attempt failed with error [${JSON.stringify(err)}]`);
       }
 
-      if (!self.reconnectId && self.options.reconnect) {
-        self.reconnectId = setTimeout(attemptReconnect(self), self.options.reconnectInterval);
+      // check if reconnect is enabled, and attempt retry if so
+      if (!pool.reconnectId && pool.options.reconnect) {
+        if (pool.state === CONNECTING && pool.options.legacyCompatMode) {
+          callback(err);
+          return;
+        }
+
+        pool.reconnectError = err;
+        pool.reconnectId = setTimeout(
+          attemptReconnect(pool, callback),
+          pool.options.reconnectInterval
+        );
+
+        return;
+      }
+
+      if (typeof callback === 'function') {
+        callback(err);
       }
 
       return;
     }
 
-    if (self.state === DESTROYED || self.state === DESTROYING) {
-      removeConnection(self, connection);
-      return connection.destroy();
+    // the pool might have been closed since we started creating the connection
+    if (pool.state === DESTROYED || pool.state === DESTROYING) {
+      if (typeof callback === 'function') {
+        callback(new MongoError('Pool was destroyed after connection creation'));
+      }
+
+      connection.destroy();
+      return;
     }
 
-    connection.on('error', self._connectionErrorHandler);
-    connection.on('close', self._connectionCloseHandler);
-    connection.on('timeout', self._connectionTimeoutHandler);
-    connection.on('parseError', self._connectionParseErrorHandler);
-    connection.on('message', self._messageHandler);
+    // otherwise, connect relevant event handlers and add it to our available connections
+    connection.on('error', pool._connectionErrorHandler);
+    connection.on('close', pool._connectionCloseHandler);
+    connection.on('timeout', pool._connectionTimeoutHandler);
+    connection.on('parseError', pool._connectionParseErrorHandler);
+    connection.on('message', pool._messageHandler);
 
-    if (self.state === DESTROYED || self.state === DESTROYING) {
-      return connection.destroy();
+    pool.availableConnections.push(connection);
+
+    // if a callback was provided, return the connection
+    if (typeof callback === 'function') {
+      callback(null, connection);
     }
 
-    // Remove the connection from the connectingConnections list
-    removeConnection(self, connection);
-
-    // Handle error
-    if (err) {
-      return connection.destroy();
-    }
-
-    // Push to available
-    self.availableConnections.push(connection);
-    // Execute any work waiting
-    _execute(self)();
+    // immediately execute any waiting work
+    _execute(pool)();
   });
 }
 
@@ -1086,6 +1078,12 @@ function _execute(self) {
       if (self.availableConnections.length === 0) {
         // Flush any monitoring operations
         flushMonitoringOperations(self.queue);
+
+        // Try to create a new connection to execute stuck operation
+        if (totalConnections < self.options.size && self.queue.length > 0) {
+          createConnection(self);
+        }
+
         break;
       }
 
@@ -1146,14 +1144,11 @@ function _execute(self) {
           // Attempt to grow the pool if it's not yet maxsize
           if (totalConnections < self.options.size && self.queue.length > 0) {
             // Create a new connection
-            _createConnection(self);
+            createConnection(self);
           }
 
           // Re-execute the operation
-          setTimeout(function() {
-            _execute(self)();
-          }, 10);
-
+          setTimeout(() => _execute(self)(), 10);
           break;
         }
       }
@@ -1166,7 +1161,7 @@ function _execute(self) {
           // Lets put the workItem back on the list
           self.queue.unshift(workItem);
           // Create a new connection
-          _createConnection(self);
+          createConnection(self);
           // Break from the loop
           break;
         }
