@@ -68,7 +68,7 @@ class CoreCursor extends Readable {
    * @param {string} ns The MongoDB fully qualified namespace (ex: db1.collection1)
    * @param {{object}|Long} cmd The selector (can be a command or a cursorId)
    * @param {object} [options=null] Optional settings.
-   * @param {object} [options.batchSize=1000] Batchsize for the operation
+   * @param {object} [options.batchSize=1000] The number of documents to return per batch. See {@link https://docs.mongodb.com/manual/reference/command/find/| find command documentation} and {@link https://docs.mongodb.com/manual/reference/command/aggregate|aggregation documentation}.
    * @param {array} [options.documents=[]] Initial documents list for cursor
    * @param {object} [options.transforms=null] Transform methods for the cursor results
    * @param {function} [options.transforms.query] Transform the value returned from the initial query
@@ -435,7 +435,7 @@ class CoreCursor extends Readable {
           return;
         }
 
-        cursor._next(callback);
+        this._initializeCursor(callback);
       });
 
       return;
