@@ -251,6 +251,10 @@ function isSDAMUnrecoverableError(error) {
   return false;
 }
 
+function isNetworkTimeoutError(err) {
+  return err instanceof MongoNetworkError && err.message.match(/timed out/);
+}
+
 module.exports = {
   MongoError,
   MongoNetworkError,
@@ -261,5 +265,6 @@ module.exports = {
   mongoErrorContextSymbol,
   isRetryableError,
   isSDAMUnrecoverableError,
-  isNodeShuttingDownError
+  isNodeShuttingDownError,
+  isNetworkTimeoutError
 };
