@@ -13,6 +13,9 @@ const MongoNetworkError = require('../error').MongoNetworkError;
 const collationNotSupported = require('../utils').collationNotSupported;
 const debugOptions = require('../connection/utils').debugOptions;
 const isSDAMUnrecoverableError = require('../error').isSDAMUnrecoverableError;
+const isNetworkTimeoutError = require('../error').isNetworkTimeoutError;
+const isNodeShuttingDownError = require('../error').isNodeShuttingDownError;
+const maxWireVersion = require('../utils').maxWireVersion;
 const makeStateMachine = require('../utils').makeStateMachine;
 const common = require('./common');
 
@@ -139,15 +142,7 @@ class Server extends EventEmitter {
       this.s.pool.clear();
     });
 
-    this[kMonitor].on('resetServer', error => {
-      // Revert to an `Unknown` state by emitting a default description with no isMaster, and the
-      // error from the heartbeat attempt
-      this.emit(
-        'descriptionReceived',
-        new ServerDescription(this.description.address, null, { error })
-      );
-    });
-
+    this[kMonitor].on('resetServer', error => markServerUnknown(this, error));
     this[kMonitor].on('serverHeartbeatSucceeded', event => {
       this.emit(
         'descriptionReceived',
@@ -278,21 +273,12 @@ class Server extends EventEmitter {
     }
 
     this.s.pool.withConnection((err, conn, cb) => {
-      if (err) return cb(err);
+      if (err) {
+        markServerUnknown(this, err);
+        return cb(err);
+      }
 
-      conn.command(ns, cmd, options, (err, result) => {
-        if (err) {
-          if (options.session && err instanceof MongoNetworkError) {
-            options.session.serverSession.isDirty = true;
-          }
-
-          if (isSDAMUnrecoverableError(err)) {
-            this.emit('error', err);
-          }
-        }
-
-        cb(err, result);
-      });
+      conn.command(ns, cmd, options, makeOperationHandler(this, options, cb));
     }, callback);
   }
 
@@ -311,21 +297,12 @@ class Server extends EventEmitter {
     }
 
     this.s.pool.withConnection((err, conn, cb) => {
-      if (err) return cb(err);
+      if (err) {
+        markServerUnknown(this, err);
+        return cb(err);
+      }
 
-      conn.query(ns, cmd, cursorState, options, (err, result) => {
-        if (err) {
-          if (options.session && err instanceof MongoNetworkError) {
-            options.session.serverSession.isDirty = true;
-          }
-
-          if (isSDAMUnrecoverableError(err)) {
-            this.emit('error', err);
-          }
-        }
-
-        cb(err, result);
-      });
+      conn.query(ns, cmd, cursorState, options, makeOperationHandler(this, options, cb));
     }, callback);
   }
 
@@ -344,21 +321,12 @@ class Server extends EventEmitter {
     }
 
     this.s.pool.withConnection((err, conn, cb) => {
-      if (err) return cb(err);
+      if (err) {
+        markServerUnknown(this, err);
+        return cb(err);
+      }
 
-      conn.getMore(ns, cursorState, batchSize, options, (err, result) => {
-        if (err) {
-          if (options.session && err instanceof MongoNetworkError) {
-            options.session.serverSession.isDirty = true;
-          }
-
-          if (isSDAMUnrecoverableError(err)) {
-            this.emit('error', err);
-          }
-        }
-
-        cb(err, result);
-      });
+      conn.getMore(ns, cursorState, batchSize, options, makeOperationHandler(this, options, cb));
     }, callback);
   }
 
@@ -379,15 +347,12 @@ class Server extends EventEmitter {
     }
 
     this.s.pool.withConnection((err, conn, cb) => {
-      if (err) return cb(err);
+      if (err) {
+        markServerUnknown(this, err);
+        return cb(err);
+      }
 
-      conn.killCursors(ns, cursorState, (err, result) => {
-        if (err && isSDAMUnrecoverableError(err)) {
-          this.emit('error', err);
-        }
-
-        cb(err, result);
-      });
+      conn.killCursors(ns, cursorState, makeOperationHandler(this, null, cb));
     }, callback);
   }
 
@@ -481,22 +446,46 @@ function executeWriteOperation(args, options, callback) {
   }
 
   server.s.pool.withConnection((err, conn, cb) => {
-    if (err) return cb(err);
+    if (err) {
+      markServerUnknown(server, err);
+      return cb(err);
+    }
 
-    conn[op](ns, ops, options, (err, result) => {
-      if (err) {
-        if (options.session && err instanceof MongoNetworkError) {
+    conn[op](ns, ops, options, makeOperationHandler(server, options, cb));
+  }, callback);
+}
+
+function markServerUnknown(server, error) {
+  server.emit(
+    'descriptionReceived',
+    new ServerDescription(server.description.address, null, { error })
+  );
+}
+
+function makeOperationHandler(server, options, callback) {
+  return function handleOperationResult(err, result) {
+    if (err) {
+      if (err instanceof MongoNetworkError) {
+        if (options && options.session) {
           options.session.serverSession.isDirty = true;
         }
 
-        if (isSDAMUnrecoverableError(err)) {
-          server.emit('error', err);
+        if (!isNetworkTimeoutError(err)) {
+          markServerUnknown(server, err);
+          server.s.pool.clear();
+        }
+      } else if (isSDAMUnrecoverableError(err)) {
+        if (maxWireVersion(server) <= 7 || isNodeShuttingDownError(err)) {
+          server.s.pool.clear();
         }
-      }
 
-      cb(err, result);
-    });
-  }, callback);
+        markServerUnknown(server, err);
+        process.nextTick(() => server.requestCheck());
+      }
+    }
+
+    callback(err, result);
+  };
 }
 
 module.exports = {
