@@ -15,8 +15,6 @@ const deprecate = require('util').deprecate;
 const BSON = require('../connection/utils').retrieveBSON();
 const createCompressionInfo = require('../topologies/shared').createCompressionInfo;
 const isRetryableError = require('../error').isRetryableError;
-const isNodeShuttingDownError = require('../error').isNodeShuttingDownError;
-const maxWireVersion = require('../utils').maxWireVersion;
 const ClientSession = require('../sessions').ClientSession;
 const MongoError = require('../error').MongoError;
 const MongoServerSelectionError = require('../error').MongoServerSelectionError;
@@ -37,7 +35,6 @@ const clearAndRemoveTimerFrom = common.clearAndRemoveTimerFrom;
 const serverSelection = require('./server_selection');
 const readPreferenceServerSelector = serverSelection.readPreferenceServerSelector;
 const writableServerSelector = serverSelection.writableServerSelector;
-// const selectServers = serverSelection.selectServers;
 
 // Global state
 let globalTopologyCounter = 0;
@@ -56,7 +53,7 @@ const SERVER_RELAY_EVENTS = [
 ].concat(CMAP_EVENT_NAMES);
 
 // all events we listen to from `Server` instances
-const LOCAL_SERVER_EVENTS = ['error', 'connect', 'descriptionReceived', 'close', 'ended'];
+const LOCAL_SERVER_EVENTS = ['connect', 'descriptionReceived', 'close', 'ended'];
 
 const STATE_CLOSING = common.STATE_CLOSING;
 const STATE_CLOSED = common.STATE_CLOSED;
@@ -280,7 +277,7 @@ class Topology extends EventEmitter {
 
     translateReadPreference(options);
     const readPreference = options.readPreference || ReadPreference.primary;
-    this.selectServer(readPreferenceServerSelector(readPreference), options, (err, server) => {
+    this.selectServer(readPreferenceServerSelector(readPreference), options, err => {
       if (err) {
         this.close();
 
@@ -293,28 +290,11 @@ class Topology extends EventEmitter {
         return;
       }
 
-      const errorHandler = err => {
-        stateTransition(this, STATE_CLOSED);
-        server.removeListener('connect', connectHandler);
-        if (typeof callback === 'function') callback(err, null);
-      };
+      stateTransition(this, STATE_CONNECTED);
+      this.emit('open', err, this);
+      this.emit('connect', this);
 
-      const connectHandler = (_, err) => {
-        stateTransition(this, STATE_CONNECTED);
-        server.removeListener('error', errorHandler);
-        this.emit('open', err, this);
-        this.emit('connect', this);
-
-        if (typeof callback === 'function') callback(err, this);
-      };
-
-      if (server.s.state === STATE_CONNECTING) {
-        server.once('error', errorHandler);
-        server.once('connect', connectHandler);
-        return;
-      }
-
-      connectHandler();
+      if (typeof callback === 'function') callback(err, this);
     });
   }
 
@@ -332,7 +312,7 @@ class Topology extends EventEmitter {
     }
 
     options = options || {};
-    if (this.s.state === STATE_CLOSED) {
+    if (this.s.state === STATE_CLOSED || this.s.state === STATE_CLOSING) {
       if (typeof callback === 'function') {
         callback();
       }
@@ -340,6 +320,8 @@ class Topology extends EventEmitter {
       return;
     }
 
+    stateTransition(this, STATE_CLOSING);
+
     drainWaitQueue(this[kWaitQueue], new MongoError('Topology closed'));
     drainTimerQueue(this.s.connectionTimers);
 
@@ -356,8 +338,6 @@ class Topology extends EventEmitter {
       delete this.s.detectTopologyDescriptionChange;
     }
 
-    stateTransition(this, STATE_CLOSING);
-
     this.s.sessions.forEach(session => session.endSession());
     this.s.sessionPool.endAllPooledSessions(() => {
       eachAsync(
@@ -398,6 +378,8 @@ class Topology extends EventEmitter {
         let readPreference;
         if (selector instanceof ReadPreference) {
           readPreference = selector;
+        } else if (typeof selector === 'string') {
+          readPreference = new ReadPreference(selector);
         } else {
           translateReadPreference(options);
           readPreference = options.readPreference || ReadPreference.primary;
@@ -454,8 +436,7 @@ class Topology extends EventEmitter {
       }, serverSelectionTimeoutMS);
     }
 
-    // place the member at the front of the wait queue
-    this[kWaitQueue].unshift(waitQueueMember);
+    this[kWaitQueue].push(waitQueueMember);
     processWaitQueue(this);
   }
 
@@ -540,10 +521,10 @@ class Topology extends EventEmitter {
     }
 
     // If we already know all the information contained in this updated description, then
-    // we don't need to update anything or emit SDAM events
-    if (previousServerDescription && previousServerDescription.equals(serverDescription)) {
-      return;
-    }
+    // we don't need to emit SDAM events, but still need to update the description, in order
+    // to keep client-tracked attributes like last update time and round trip time up to date
+    const equalDescriptions =
+      previousServerDescription && previousServerDescription.equals(serverDescription);
 
     // first update the TopologyDescription
     this.s.description = this.s.description.update(serverDescription);
@@ -553,15 +534,17 @@ class Topology extends EventEmitter {
     }
 
     // emit monitoring events for this change
-    this.emit(
-      'serverDescriptionChanged',
-      new events.ServerDescriptionChangedEvent(
-        this.s.id,
-        serverDescription.address,
-        previousServerDescription,
-        this.s.description.servers.get(serverDescription.address)
-      )
-    );
+    if (!equalDescriptions) {
+      this.emit(
+        'serverDescriptionChanged',
+        new events.ServerDescriptionChangedEvent(
+          this.s.id,
+          serverDescription.address,
+          previousServerDescription,
+          this.s.description.servers.get(serverDescription.address)
+        )
+      );
+    }
 
     // update server list from updated descriptions
     updateServers(this, serverDescription);
@@ -571,14 +554,16 @@ class Topology extends EventEmitter {
       processWaitQueue(this);
     }
 
-    this.emit(
-      'topologyDescriptionChanged',
-      new events.TopologyDescriptionChangedEvent(
-        this.s.id,
-        previousTopologyDescription,
-        this.s.description
-      )
-    );
+    if (!equalDescriptions) {
+      this.emit(
+        'topologyDescriptionChanged',
+        new events.TopologyDescriptionChangedEvent(
+          this.s.id,
+          previousTopologyDescription,
+          this.s.description
+        )
+      );
+    }
   }
 
   auth(credentials, callback) {
@@ -841,7 +826,6 @@ function createAndConnectServer(topology, serverDescription, connectDelay) {
   relayEvents(server, topology, SERVER_RELAY_EVENTS);
 
   server.on('descriptionReceived', topology.serverUpdateHandler.bind(topology));
-  server.on('error', serverErrorEventHandler(server, topology));
 
   if (connectDelay) {
     const connectTimer = setTimeout(() => {
@@ -902,21 +886,6 @@ function updateServers(topology, incomingServerDescription) {
   }
 }
 
-function serverErrorEventHandler(server, topology) {
-  return function(err) {
-    if (topology.s.state === STATE_CLOSING || topology.s.state === STATE_CLOSED) {
-      return;
-    }
-
-    if (maxWireVersion(server) >= 8 && !isNodeShuttingDownError(err)) {
-      resetServerState(server, err);
-      return;
-    }
-
-    resetServerState(server, err, { clearPool: true });
-  };
-}
-
 function executeWriteOperation(args, options, callback) {
   if (typeof options === 'function') (callback = options), (options = {});
   options = options || {};
@@ -970,30 +939,6 @@ function executeWriteOperation(args, options, callback) {
   });
 }
 
-/**
- * Resets the internal state of this server to `Unknown` by simulating an empty ismaster
- *
- * @private
- * @param {Server} server
- * @param {MongoError} error The error that caused the state reset
- * @param {object} [options] Optional settings
- * @param {boolean} [options.clearPool=false] Pool should be cleared out on state reset
- */
-function resetServerState(server, error, options) {
-  options = Object.assign({}, { clearPool: false }, options);
-
-  if (options.clearPool && server.s.pool) {
-    server.s.pool.clear();
-  }
-
-  server.emit(
-    'descriptionReceived',
-    new ServerDescription(server.description.address, null, { error })
-  );
-
-  process.nextTick(() => server.requestCheck());
-}
-
 function translateReadPreference(options) {
   if (options.readPreference == null) {
     return;
@@ -1040,7 +985,7 @@ function srvPollingHandler(topology) {
 
 function drainWaitQueue(queue, err) {
   while (queue.length) {
-    const waitQueueMember = queue.pop();
+    const waitQueueMember = queue.shift();
     clearTimeout(waitQueueMember.timer);
     if (!waitQueueMember[kCancelled]) {
       waitQueueMember.callback(err);
@@ -1054,9 +999,9 @@ function processWaitQueue(topology) {
     return;
   }
 
-  const isSharded = topology.description.type === TopologyType.Sharded;
   const serverDescriptions = Array.from(topology.description.servers.values());
-  for (let i = 0; i < topology[kWaitQueue].length; ++i) {
+  const membersToProcess = topology[kWaitQueue].length;
+  for (let i = 0; i < membersToProcess && topology[kWaitQueue].length; ++i) {
     const waitQueueMember = topology[kWaitQueue].shift();
     if (waitQueueMember[kCancelled]) {
       continue;
@@ -1071,17 +1016,18 @@ function processWaitQueue(topology) {
     } catch (e) {
       clearTimeout(waitQueueMember.timer);
       waitQueueMember.callback(e);
-      break;
+      continue;
     }
 
     if (selectedDescriptions.length === 0) {
       topology[kWaitQueue].push(waitQueueMember);
-      break;
+      continue;
     }
 
     const selectedServerDescription = randomSelection(selectedDescriptions);
     const selectedServer = topology.s.servers.get(selectedServerDescription.address);
     const transaction = waitQueueMember.transaction;
+    const isSharded = topology.description.type === TopologyType.Sharded;
     if (isSharded && transaction && transaction.isActive) {
       transaction.pinServer(selectedServer);
     }
