@@ -222,10 +222,8 @@ class ClientSession extends EventEmitter {
     }
 
     return new Promise((resolve, reject) => {
-      endTransaction(
-        this,
-        'commitTransaction',
-        (err, reply) => (err ? reject(err) : resolve(reply))
+      endTransaction(this, 'commitTransaction', (err, reply) =>
+        err ? reject(err) : resolve(reply)
       );
     });
   }
@@ -243,10 +241,8 @@ class ClientSession extends EventEmitter {
     }
 
     return new Promise((resolve, reject) => {
-      endTransaction(
-        this,
-        'abortTransaction',
-        (err, reply) => (err ? reject(err) : resolve(reply))
+      endTransaction(this, 'abortTransaction', (err, reply) =>
+        err ? reject(err) : resolve(reply)
       );
     });
   }
@@ -592,10 +588,23 @@ class ServerSessionPool {
    * Ends all sessions in the session pool.
    * @ignore
    */
-  endAllPooledSessions() {
+  endAllPooledSessions(callback) {
     if (this.sessions.length) {
-      this.topology.endSessions(this.sessions.map(session => session.id));
-      this.sessions = [];
+      this.topology.endSessions(
+        this.sessions.map(session => session.id),
+        () => {
+          this.sessions = [];
+          if (typeof callback === 'function') {
+            callback();
+          }
+        }
+      );
+
+      return;
+    }
+
+    if (typeof callback === 'function') {
+      callback();
     }
   }
 
@@ -674,6 +683,7 @@ function commandSupportsReadConcern(command, options) {
 /**
  * Optionally decorate a command with sessions specific keys
  *
+ * @ignore
  * @param {ClientSession} session the session tracking transaction state
  * @param {Object} command the command to decorate
  * @param {Object} topology the topology for tracking the cluster time
