@@ -8,16 +8,15 @@ const Logger = require('../connection/logger');
 const retrieveBSON = require('../connection/utils').retrieveBSON;
 const MongoError = require('../error').MongoError;
 const Server = require('./server');
-const clone = require('./shared').clone;
 const diff = require('./shared').diff;
 const cloneOptions = require('./shared').cloneOptions;
-const createClientInfo = require('./shared').createClientInfo;
 const SessionMixins = require('./shared').SessionMixins;
 const isRetryableWritesSupported = require('./shared').isRetryableWritesSupported;
 const relayEvents = require('../utils').relayEvents;
 const isRetryableError = require('../error').isRetryableError;
 const BSON = retrieveBSON();
 const getMMAPError = require('./shared').getMMAPError;
+const makeClientMetadata = require('../utils').makeClientMetadata;
 
 /**
  * @fileOverview The **Mongos** class is a class that represents a Mongos Proxy topology and is
@@ -30,14 +29,15 @@ var DISCONNECTED = 'disconnected';
 var CONNECTING = 'connecting';
 var CONNECTED = 'connected';
 var UNREFERENCED = 'unreferenced';
+var DESTROYING = 'destroying';
 var DESTROYED = 'destroyed';
 
 function stateTransition(self, newState) {
   var legalTransitions = {
-    disconnected: [CONNECTING, DESTROYED, DISCONNECTED],
-    connecting: [CONNECTING, DESTROYED, CONNECTED, DISCONNECTED],
-    connected: [CONNECTED, DISCONNECTED, DESTROYED, UNREFERENCED],
-    unreferenced: [UNREFERENCED, DESTROYED],
+    disconnected: [CONNECTING, DESTROYING, DESTROYED, DISCONNECTED],
+    connecting: [CONNECTING, DESTROYING, DESTROYED, CONNECTED, DISCONNECTED],
+    connected: [CONNECTED, DISCONNECTED, DESTROYING, DESTROYED, UNREFERENCED],
+    unreferenced: [UNREFERENCED, DESTROYING, DESTROYED],
     destroyed: [DESTROYED]
   };
 
@@ -46,9 +46,9 @@ function stateTransition(self, newState) {
   if (legalStates && legalStates.indexOf(newState) !== -1) {
     self.state = newState;
   } else {
-    self.logger.error(
+    self.s.logger.error(
       f(
-        'Pool with id [%s] failed attempted illegal state transition from [%s] to [%s] only following state allowed [%s]',
+        'Mongos with id [%s] failed attempted illegal state transition from [%s] to [%s] only following state allowed [%s]',
         self.id,
         self.state,
         newState,
@@ -115,7 +115,7 @@ var Mongos = function(seedlist, options) {
 
   // Internal state
   this.s = {
-    options: Object.assign({}, options),
+    options: Object.assign({ metadata: makeClientMetadata(options) }, options),
     // BSON instance
     bson:
       options.bson ||
@@ -152,14 +152,9 @@ var Mongos = function(seedlist, options) {
     // Are we running in debug mode
     debug: typeof options.debug === 'boolean' ? options.debug : false,
     // localThresholdMS
-    localThresholdMS: options.localThresholdMS || 15,
-    // Client info
-    clientInfo: createClientInfo(options)
+    localThresholdMS: options.localThresholdMS || 15
   };
 
-  // Set the client info
-  this.s.options.clientInfo = createClientInfo(options);
-
   // Log info warning if the socketTimeout < haInterval as it will cause
   // a lot of recycled connections to happen.
   if (
@@ -264,8 +259,7 @@ Mongos.prototype.connect = function(options) {
       Object.assign({}, self.s.options, x, options, {
         reconnect: false,
         monitoring: false,
-        parent: self,
-        clientInfo: clone(self.s.clientInfo)
+        parent: self
       })
     );
 
@@ -292,7 +286,10 @@ Mongos.prototype.auth = function(credentials, callback) {
 
 function handleEvent(self) {
   return function() {
-    if (self.state === DESTROYED) return;
+    if (self.state === DESTROYED || self.state === DESTROYING) {
+      return;
+    }
+
     // Move to list of disconnectedProxies
     moveServerFrom(self.connectedProxies, self.disconnectedProxies, this);
     // Emit the initial topology
@@ -366,7 +363,8 @@ function handleInitialConnectEvent(self, event) {
           self.s.logger.warn(f(message, _this.name));
         }
 
-        // This is not a mongos proxy, remove it completely
+        // This is not a mongos proxy, destroy and remove it completely
+        _this.destroy(true);
         removeProxyFrom(self.connectingProxies, _this);
         // Emit the left event
         self.emit('left', 'server', _this);
@@ -445,10 +443,9 @@ function connectProxies(self, servers) {
       server.connect(self.s.connectOptions);
     }, timeoutInterval);
   }
+
   // Start all the servers
-  while (servers.length > 0) {
-    connect(servers.shift(), timeoutInterval++);
-  }
+  servers.forEach(server => connect(server, timeoutInterval++));
 }
 
 function pickProxy(self, session) {
@@ -541,14 +538,14 @@ function reconnectProxies(self, proxies, callback) {
       count = count - 1;
 
       // Destroyed
-      if (self.state === DESTROYED || self.state === UNREFERENCED) {
+      if (self.state === DESTROYED || self.state === DESTROYING || self.state === UNREFERENCED) {
         moveServerFrom(self.connectingProxies, self.disconnectedProxies, _self);
         return this.destroy();
       }
 
       if (event === 'connect') {
         // Destroyed
-        if (self.state === DESTROYED || self.state === UNREFERENCED) {
+        if (self.state === DESTROYED || self.state === DESTROYING || self.state === UNREFERENCED) {
           moveServerFrom(self.connectingProxies, self.disconnectedProxies, _self);
           return _self.destroy();
         }
@@ -592,7 +589,7 @@ function reconnectProxies(self, proxies, callback) {
   function execute(_server, i) {
     setTimeout(function() {
       // Destroyed
-      if (self.state === DESTROYED || self.state === UNREFERENCED) {
+      if (self.state === DESTROYED || self.state === DESTROYING || self.state === UNREFERENCED) {
         return;
       }
 
@@ -603,12 +600,11 @@ function reconnectProxies(self, proxies, callback) {
           port: parseInt(_server.name.split(':')[1], 10),
           reconnect: false,
           monitoring: false,
-          parent: self,
-          clientInfo: clone(self.s.clientInfo)
+          parent: self
         })
       );
 
-      destroyServer(_server);
+      destroyServer(_server, { force: true });
       removeProxyFrom(self.disconnectedProxies, _server);
 
       // Relay the server description change
@@ -645,9 +641,17 @@ function reconnectProxies(self, proxies, callback) {
 function topologyMonitor(self, options) {
   options = options || {};
 
+  // no need to set up the monitor if we're already closed
+  if (self.state === DESTROYED || self.state === DESTROYING || self.state === UNREFERENCED) {
+    return;
+  }
+
   // Set momitoring timeout
   self.haTimeoutId = setTimeout(function() {
-    if (self.state === DESTROYED || self.state === UNREFERENCED) return;
+    if (self.state === DESTROYED || self.state === DESTROYING || self.state === UNREFERENCED) {
+      return;
+    }
+
     // If we have a primary and a disconnect handler, execute
     // buffered operations
     if (self.isConnected() && self.s.disconnectHandler) {
@@ -678,7 +682,11 @@ function topologyMonitor(self, options) {
           socketTimeout: self.s.options.connectionTimeout || 2000
         },
         function(err, r) {
-          if (self.state === DESTROYED || self.state === UNREFERENCED) {
+          if (
+            self.state === DESTROYED ||
+            self.state === DESTROYING ||
+            self.state === UNREFERENCED
+          ) {
             // Move from connectingProxies
             moveServerFrom(self.connectedProxies, self.disconnectedProxies, _server);
             _server.destroy();
@@ -727,7 +735,9 @@ function topologyMonitor(self, options) {
 
       // Attempt to connect to any unknown servers
       return reconnectProxies(self, self.disconnectedProxies, function() {
-        if (self.state === DESTROYED || self.state === UNREFERENCED) return;
+        if (self.state === DESTROYED || self.state === DESTROYING || self.state === UNREFERENCED) {
+          return;
+        }
 
         // Are we connected ? emit connect event
         if (self.state === CONNECTING && options.firstConnect) {
@@ -751,11 +761,24 @@ function topologyMonitor(self, options) {
         count = count - 1;
 
         if (count === 0) {
-          if (self.state === DESTROYED || self.state === UNREFERENCED) return;
+          if (
+            self.state === DESTROYED ||
+            self.state === DESTROYING ||
+            self.state === UNREFERENCED
+          ) {
+            return;
+          }
 
           // Attempt to connect to any unknown servers
           reconnectProxies(self, self.disconnectedProxies, function() {
-            if (self.state === DESTROYED || self.state === UNREFERENCED) return;
+            if (
+              self.state === DESTROYED ||
+              self.state === DESTROYING ||
+              self.state === UNREFERENCED
+            ) {
+              return;
+            }
+
             // Perform topology monitor
             topologyMonitor(self);
           });
@@ -796,6 +819,14 @@ Mongos.prototype.unref = function() {
  * @method
  */
 Mongos.prototype.destroy = function(options, callback) {
+  if (typeof options === 'function') {
+    callback = options;
+    options = {};
+  }
+
+  options = options || {};
+
+  stateTransition(this, DESTROYING);
   if (this.haTimeoutId) {
     clearTimeout(this.haTimeoutId);
   }
@@ -929,7 +960,9 @@ Mongos.prototype.insert = function(ns, ops, options, callback) {
     (callback = options), (options = {}), (options = options || {});
   }
 
-  if (this.state === DESTROYED) return callback(new MongoError(f('topology was destroyed')));
+  if (this.state === DESTROYED) {
+    return callback(new MongoError(f('topology was destroyed')));
+  }
 
   // Not connected but we have a disconnecthandler
   if (!this.isConnected() && this.s.disconnectHandler != null) {
@@ -963,7 +996,9 @@ Mongos.prototype.update = function(ns, ops, options, callback) {
     (callback = options), (options = {}), (options = options || {});
   }
 
-  if (this.state === DESTROYED) return callback(new MongoError(f('topology was destroyed')));
+  if (this.state === DESTROYED) {
+    return callback(new MongoError(f('topology was destroyed')));
+  }
 
   // Not connected but we have a disconnecthandler
   if (!this.isConnected() && this.s.disconnectHandler != null) {
@@ -997,7 +1032,9 @@ Mongos.prototype.remove = function(ns, ops, options, callback) {
     (callback = options), (options = {}), (options = options || {});
   }
 
-  if (this.state === DESTROYED) return callback(new MongoError(f('topology was destroyed')));
+  if (this.state === DESTROYED) {
+    return callback(new MongoError(f('topology was destroyed')));
+  }
 
   // Not connected but we have a disconnecthandler
   if (!this.isConnected() && this.s.disconnectHandler != null) {
@@ -1036,7 +1073,10 @@ Mongos.prototype.command = function(ns, cmd, options, callback) {
     (callback = options), (options = {}), (options = options || {});
   }
 
-  if (this.state === DESTROYED) return callback(new MongoError(f('topology was destroyed')));
+  if (this.state === DESTROYED) {
+    return callback(new MongoError(f('topology was destroyed')));
+  }
+
   var self = this;
 
   // Pick a proxy
