@@ -10,10 +10,8 @@ const Logger = require('../connection/logger');
 const MongoError = require('../error').MongoError;
 const Server = require('./server');
 const ReplSetState = require('./replset_state');
-const clone = require('./shared').clone;
 const Timeout = require('./shared').Timeout;
 const Interval = require('./shared').Interval;
-const createClientInfo = require('./shared').createClientInfo;
 const SessionMixins = require('./shared').SessionMixins;
 const isRetryableWritesSupported = require('./shared').isRetryableWritesSupported;
 const relayEvents = require('../utils').relayEvents;
@@ -21,6 +19,7 @@ const isRetryableError = require('../error').isRetryableError;
 const BSON = retrieveBSON();
 const calculateDurationInMs = require('../utils').calculateDurationInMs;
 const getMMAPError = require('./shared').getMMAPError;
+const makeClientMetadata = require('../utils').makeClientMetadata;
 
 //
 // States
@@ -140,7 +139,7 @@ var ReplSet = function(seedlist, options) {
 
   // Internal state
   this.s = {
-    options: Object.assign({}, options),
+    options: Object.assign({ metadata: makeClientMetadata(options) }, options),
     // BSON instance
     bson:
       options.bson ||
@@ -187,9 +186,7 @@ var ReplSet = function(seedlist, options) {
     // Connect function options passed in
     connectOptions: {},
     // Are we running in debug mode
-    debug: typeof options.debug === 'boolean' ? options.debug : false,
-    // Client info
-    clientInfo: createClientInfo(options)
+    debug: typeof options.debug === 'boolean' ? options.debug : false
   };
 
   // Add handler for topology change
@@ -277,27 +274,34 @@ function rexecuteOperations(self) {
 }
 
 function connectNewServers(self, servers, callback) {
+  // No new servers
+  if (servers.length === 0) {
+    return callback();
+  }
+
   // Count lefts
   var count = servers.length;
   var error = null;
 
+  function done() {
+    count = count - 1;
+    if (count === 0) {
+      callback(error);
+    }
+  }
+
   // Handle events
   var _handleEvent = function(self, event) {
     return function(err) {
       var _self = this;
-      count = count - 1;
 
       // Destroyed
       if (self.state === DESTROYED || self.state === UNREFERENCED) {
-        return this.destroy({ force: true });
+        this.destroy({ force: true });
+        return done();
       }
 
       if (event === 'connect') {
-        // Destroyed
-        if (self.state === DESTROYED || self.state === UNREFERENCED) {
-          return _self.destroy({ force: true });
-        }
-
         // Update the state
         var result = self.s.replicaSetState.update(_self);
         // Update the state with the new server
@@ -332,17 +336,10 @@ function connectNewServers(self, servers, callback) {
 
       // Rexecute any stalled operation
       rexecuteOperations(self);
-
-      // Are we done finish up callback
-      if (count === 0) {
-        callback(error);
-      }
+      done();
     };
   };
 
-  // No new servers
-  if (count === 0) return callback();
-
   // Execute method
   function execute(_server, i) {
     setTimeout(function() {
@@ -351,6 +348,17 @@ function connectNewServers(self, servers, callback) {
         return;
       }
 
+      // remove existing connecting server if it's failed to connect, otherwise
+      // wait for that server to connect
+      const existingServerIdx = self.s.connectingServers.findIndex(s => s.name === _server);
+      if (existingServerIdx >= 0) {
+        const connectingServer = self.s.connectingServers[existingServerIdx];
+        connectingServer.destroy({ force: true });
+
+        self.s.connectingServers.splice(existingServerIdx, 1);
+        return done();
+      }
+
       // Create a new server instance
       var server = new Server(
         Object.assign({}, self.s.options, {
@@ -358,8 +366,7 @@ function connectNewServers(self, servers, callback) {
           port: parseInt(_server.split(':')[1], 10),
           reconnect: false,
           monitoring: false,
-          parent: self,
-          clientInfo: clone(self.s.clientInfo)
+          parent: self
         })
       );
 
@@ -378,6 +385,7 @@ function connectNewServers(self, servers, callback) {
       // Command Monitoring events
       relayEvents(server, self, ['commandStarted', 'commandSucceeded', 'commandFailed']);
 
+      self.s.connectingServers.push(server);
       server.connect(self.s.connectOptions);
     }, i);
   }
@@ -906,8 +914,7 @@ ReplSet.prototype.connect = function(options) {
       Object.assign({}, self.s.options, x, options, {
         reconnect: false,
         monitoring: false,
-        parent: self,
-        clientInfo: clone(self.s.clientInfo)
+        parent: self
       })
     );
   });
@@ -951,6 +958,11 @@ ReplSet.prototype.auth = function(credentials, callback) {
  * @method
  */
 ReplSet.prototype.destroy = function(options, callback) {
+  if (typeof options === 'function') {
+    callback = options;
+    options = {};
+  }
+
   options = options || {};
 
   let destroyCount = this.s.connectingServers.length + 1; // +1 for the callback from `replicaSetState.destroy`
@@ -1202,6 +1214,7 @@ function executeWriteOperation(args, options, callback) {
 
     // Per SDAM, remove primary from replicaset
     if (self.s.replicaSetState.primary) {
+      self.s.replicaSetState.primary.destroy();
       self.s.replicaSetState.remove(self.s.replicaSetState.primary, { force: true });
     }
 
@@ -1363,6 +1376,7 @@ ReplSet.prototype.command = function(ns, cmd, options, callback) {
 
     // Per SDAM, remove primary from replicaset
     if (this.s.replicaSetState.primary) {
+      this.s.replicaSetState.primary.destroy();
       this.s.replicaSetState.remove(this.s.replicaSetState.primary, { force: true });
     }
 
