@@ -13,13 +13,13 @@ var inherits = require('util').inherits,
   wireProtocol = require('../wireprotocol'),
   CoreCursor = require('../cursor').CoreCursor,
   sdam = require('./shared'),
-  createClientInfo = require('./shared').createClientInfo,
   createCompressionInfo = require('./shared').createCompressionInfo,
   resolveClusterTime = require('./shared').resolveClusterTime,
   SessionMixins = require('./shared').SessionMixins,
   relayEvents = require('../utils').relayEvents;
 
 const collationNotSupported = require('../utils').collationNotSupported;
+const makeClientMetadata = require('../utils').makeClientMetadata;
 
 // Used for filtering out fields for loggin
 var debugFields = [
@@ -55,6 +55,10 @@ var serverAccounting = false;
 var servers = {};
 var BSON = retrieveBSON();
 
+function topologyId(server) {
+  return server.s.parent == null ? server.id : server.s.parent.id;
+}
+
 /**
  * Creates a new Server instance
  * @class
@@ -116,7 +120,7 @@ var Server = function(options) {
   // Internal state
   this.s = {
     // Options
-    options: options,
+    options: Object.assign({ metadata: makeClientMetadata(options) }, options),
     // Logger
     logger: Logger('Server', options),
     // Factory overrides
@@ -151,8 +155,6 @@ var Server = function(options) {
     // Monitoring timeout
     monitoringInterval:
       typeof options.monitoringInterval === 'number' ? options.monitoringInterval : 5000,
-    // Topology id
-    topologyId: -1,
     compression: { compressors: createCompressionInfo(options) },
     // Optional parent topology
     parent: options.parent
@@ -173,8 +175,6 @@ var Server = function(options) {
   this.initialConnect = true;
   // Default type
   this._type = 'server';
-  // Set the client info
-  this.clientInfo = createClientInfo(options);
 
   // Max Stalleness values
   // last time we updated the ismaster state
@@ -210,6 +210,13 @@ Object.defineProperty(Server.prototype, 'logicalSessionTimeoutMinutes', {
   }
 });
 
+Object.defineProperty(Server.prototype, 'clientMetadata', {
+  enumerable: true,
+  get: function() {
+    return this.s.options.metadata;
+  }
+});
+
 // In single server deployments we track the clusterTime directly on the topology, however
 // in Mongos and ReplSet deployments we instead need to delegate the clusterTime up to the
 // tracking objects so we can ensure we are gossiping the maximum time received from the
@@ -486,14 +493,11 @@ Server.prototype.connect = function(options) {
 
   // Emit toplogy opening event if not in topology
   if (!self.s.inTopology) {
-    this.emit('topologyOpening', { topologyId: self.id });
+    this.emit('topologyOpening', { topologyId: topologyId(self) });
   }
 
   // Emit opening server event
-  self.emit('serverOpening', {
-    topologyId: self.s.topologyId !== -1 ? self.s.topologyId : self.id,
-    address: self.name
-  });
+  self.emit('serverOpening', { topologyId: topologyId(self), address: self.name });
 
   self.s.pool.connect();
 };
@@ -843,6 +847,11 @@ Server.prototype.destroy = function(options, callback) {
     return;
   }
 
+  if (typeof options === 'function') {
+    callback = options;
+    options = {};
+  }
+
   options = options || {};
   var self = this;
 
@@ -878,14 +887,11 @@ Server.prototype.destroy = function(options, callback) {
 
   // Emit opening server event
   if (self.listeners('serverClosed').length > 0)
-    self.emit('serverClosed', {
-      topologyId: self.s.topologyId !== -1 ? self.s.topologyId : self.id,
-      address: self.name
-    });
+    self.emit('serverClosed', { topologyId: topologyId(self), address: self.name });
 
   // Emit toplogy opening event if not in topology
   if (self.listeners('topologyClosed').length > 0 && !self.s.inTopology) {
-    self.emit('topologyClosed', { topologyId: self.id });
+    self.emit('topologyClosed', { topologyId: topologyId(self) });
   }
 
   if (self.s.logger.isDebug()) {
