@@ -100,7 +100,7 @@ function parseSrvConnectionString(uri, options, callback) {
           );
         }
 
-        Object.assign(result.query, record);
+        result.query = Object.assign({}, record, result.query);
       }
 
       // Set completed options back into the URL object.
@@ -419,6 +419,46 @@ function parseQueryString(query, options) {
   return Object.keys(result).length ? result : null;
 }
 
+/// Adds support for modern `tls` variants of out `ssl` options
+function translateTLSOptions(queryString) {
+  if (queryString.tls) {
+    queryString.ssl = queryString.tls;
+  }
+
+  if (queryString.tlsInsecure) {
+    queryString.checkServerIdentity = false;
+    queryString.sslValidate = false;
+  } else {
+    Object.assign(queryString, {
+      checkServerIdentity: queryString.tlsAllowInvalidHostnames ? false : true,
+      sslValidate: queryString.tlsAllowInvalidCertificates ? false : true
+    });
+  }
+
+  if (queryString.tlsCAFile) {
+    queryString.ssl = true;
+    queryString.sslCA = queryString.tlsCAFile;
+  }
+
+  if (queryString.tlsCertificateKeyFile) {
+    queryString.ssl = true;
+    if (queryString.tlsCertificateFile) {
+      queryString.sslCert = queryString.tlsCertificateFile;
+      queryString.sslKey = queryString.tlsCertificateKeyFile;
+    } else {
+      queryString.sslKey = queryString.tlsCertificateKeyFile;
+      queryString.sslCert = queryString.tlsCertificateKeyFile;
+    }
+  }
+
+  if (queryString.tlsCertificateKeyFilePassword) {
+    queryString.ssl = true;
+    queryString.sslPass = queryString.tlsCertificateKeyFilePassword;
+  }
+
+  return queryString;
+}
+
 /**
  * Checks a query string for invalid tls options according to the URI options spec.
  *
@@ -470,7 +510,7 @@ function assertTlsOptionsAreEqual(optionName, queryString, queryStringKeys) {
       const firstValue = queryString[optionName][0];
       queryString[optionName].forEach(tlsValue => {
         if (tlsValue !== firstValue) {
-          throw new MongoParseError('All values of ${optionName} must be the same.');
+          throw new MongoParseError(`All values of ${optionName} must be the same.`);
         }
       });
     }
@@ -534,6 +574,10 @@ function parseConnectionString(uri, options, callback) {
     if (parsedOptions.auth.username) auth.username = parsedOptions.auth.username;
     if (parsedOptions.auth.user) auth.username = parsedOptions.auth.user;
     if (parsedOptions.auth.password) auth.password = parsedOptions.auth.password;
+  } else {
+    if (parsedOptions.username) auth.username = parsedOptions.username;
+    if (parsedOptions.user) auth.username = parsedOptions.user;
+    if (parsedOptions.password) auth.password = parsedOptions.password;
   }
 
   if (cap[4].split('?')[0].indexOf('@') !== -1) {
@@ -545,14 +589,22 @@ function parseConnectionString(uri, options, callback) {
     return callback(new MongoParseError('Unescaped at-sign in authority section'));
   }
 
+  if (authorityParts[0] == null || authorityParts[0] === '') {
+    return callback(new MongoParseError('No username provided in authority section'));
+  }
+
   if (authorityParts.length > 1) {
     const authParts = authorityParts.shift().split(':');
     if (authParts.length > 2) {
       return callback(new MongoParseError('Unescaped colon in authority section'));
     }
 
-    auth.username = qs.unescape(authParts[0]);
-    auth.password = authParts[1] ? qs.unescape(authParts[1]) : null;
+    if (authParts[0] === '') {
+      return callback(new MongoParseError('Invalid empty username provided'));
+    }
+
+    if (!auth.username) auth.username = qs.unescape(authParts[0]);
+    if (!auth.password) auth.password = authParts[1] ? qs.unescape(authParts[1]) : null;
   }
 
   let hostParsingError = null;
@@ -617,8 +669,13 @@ function parseConnectionString(uri, options, callback) {
 
   if (result.auth && result.auth.db) {
     result.defaultDatabase = result.auth.db;
+  } else {
+    result.defaultDatabase = 'test';
   }
 
+  // support modern `tls` variants to SSL options
+  result.options = translateTLSOptions(result.options);
+
   try {
     applyAuthExpectations(result);
   } catch (authError) {
