@@ -224,14 +224,21 @@ var open = function(self, options, callback) {
       var chunkIndexOptions = shallowClone(writeConcern);
       chunkIndexOptions.unique = true;
       // Ensure index on chunk collection
-      chunkCollection.ensureIndex([['files_id', 1], ['n', 1]], chunkIndexOptions, function() {
-        // Open the connection
-        _open(self, writeConcern, function(err, r) {
-          if (err) return callback(err);
-          self.isOpen = true;
-          callback(err, r);
-        });
-      });
+      chunkCollection.ensureIndex(
+        [
+          ['files_id', 1],
+          ['n', 1]
+        ],
+        chunkIndexOptions,
+        function() {
+          // Open the connection
+          _open(self, writeConcern, function(err, r) {
+            if (err) return callback(err);
+            self.isOpen = true;
+            callback(err, r);
+          });
+        }
+      );
     });
   } else {
     // Open the gridstore
