@@ -3,17 +3,16 @@
 const ChangeStream = require('./change_stream');
 const Db = require('./db');
 const EventEmitter = require('events').EventEmitter;
-const executeOperation = require('./operations/execute_operation');
 const inherits = require('util').inherits;
 const MongoError = require('./core').MongoError;
 const deprecate = require('util').deprecate;
 const WriteConcern = require('./write_concern');
 const MongoDBNamespace = require('./utils').MongoDBNamespace;
 const ReadPreference = require('./core/topologies/read_preference');
-
-// Operations
-const ConnectOperation = require('./operations/connect');
-const CloseOperation = require('./operations/close');
+const maybePromise = require('./utils').maybePromise;
+const NativeTopology = require('./topologies/native_topology');
+const connect = require('./operations/connect').connect;
+const validOptions = require('./operations/connect').validOptions;
 
 /**
  * @fileOverview The **MongoClient** class is a class that allows for making Connections to MongoDB.
@@ -55,26 +54,21 @@ const CloseOperation = require('./operations/close');
  */
 
 /**
- * Configuration options for a automatic client encryption.
+ * Configuration options for drivers wrapping the node driver.
  *
- * **NOTE**: Support for client side encryption is in beta. Backwards-breaking changes may be made before the final release.
+ * @typedef {Object} DriverInfoOptions
+ * @property {string} [name] The name of the driver
+ * @property {string} [version] The version of the driver
+ * @property {string} [platform] Optional platform information
+ */
+
+/**
+ * Configuration options for drivers wrapping the node driver.
  *
- * @typedef {Object} AutoEncryptionOptions
- * @property {MongoClient} [keyVaultClient] A `MongoClient` used to fetch keys from a key vault
- * @property {string} [keyVaultNamespace] The namespace where keys are stored in the key vault
- * @property {object} [kmsProviders] Provider details for the desired Key Management Service to use for encryption
- * @property {object} [kmsProviders.aws] Optional settings for the AWS KMS provider
- * @property {string} [kmsProviders.aws.accessKeyId] The access key used for the AWS KMS provider
- * @property {string} [kmsProviders.aws.secretAccessKey] The secret access key used for the AWS KMS provider
- * @property {object} [kmsProviders.local] Optional settings for the local KMS provider
- * @property {string} [kmsProviders.local.key] The master key used to encrypt/decrypt data keys
- * @property {object} [schemaMap] A map of namespaces to a local JSON schema for encryption
- * @property {boolean} [bypassAutoEncryption] Allows the user to bypass auto encryption, maintaining implicit decryption
- * @property {object} [extraOptions] Extra options related to the mongocryptd process
- * @property {string} [extraOptions.mongocryptURI] A local process the driver communicates with to determine how to encrypt values in a command. Defaults to "mongodb://%2Fvar%2Fmongocryptd.sock" if domain sockets are available or "mongodb://localhost:27020" otherwise
- * @property {boolean} [extraOptions.mongocryptdBypassSpawn=false] If true, autoEncryption will not attempt to spawn a mongocryptd before connecting
- * @property {string} [extraOptions.mongocryptdSpawnPath] The path to the mongocryptd executable on the system
- * @property {string[]} [extraOptions.mongocryptdSpawnArgs] Command line arguments to use when auto-spawning a mongocryptd
+ * @typedef {Object} DriverInfoOptions
+ * @property {string} [name] The name of the driver
+ * @property {string} [version] The version of the driver
+ * @property {string} [platform] Optional platform information
  */
 
 /**
@@ -83,21 +77,29 @@ const CloseOperation = require('./operations/close');
  * @param {string} url The connection URI string
  * @param {object} [options] Optional settings
  * @param {number} [options.poolSize=5] The maximum size of the individual server pool
- * @param {boolean} [options.ssl=false] Enable SSL connection.
+ * @param {boolean} [options.ssl=false] Enable SSL connection. *deprecated* use `tls` variants
  * @param {boolean} [options.sslValidate=false] Validate mongod server certificate against Certificate Authority
- * @param {buffer} [options.sslCA=undefined] SSL Certificate store binary buffer
- * @param {buffer} [options.sslCert=undefined] SSL Certificate binary buffer
- * @param {buffer} [options.sslKey=undefined] SSL Key file binary buffer
- * @param {string} [options.sslPass=undefined] SSL Certificate pass phrase
- * @param {buffer} [options.sslCRL=undefined] SSL Certificate revocation list binary buffer
+ * @param {buffer} [options.sslCA=undefined] SSL Certificate store binary buffer *deprecated* use `tls` variants
+ * @param {buffer} [options.sslCert=undefined] SSL Certificate binary buffer *deprecated* use `tls` variants
+ * @param {buffer} [options.sslKey=undefined] SSL Key file binary buffer *deprecated* use `tls` variants
+ * @param {string} [options.sslPass=undefined] SSL Certificate pass phrase *deprecated* use `tls` variants
+ * @param {buffer} [options.sslCRL=undefined] SSL Certificate revocation list binary buffer *deprecated* use `tls` variants
+ * @param {boolean|function} [options.checkServerIdentity=true] Ensure we check server identify during SSL, set to false to disable checking. Only works for Node 0.12.x or higher. You can pass in a boolean or your own checkServerIdentity override function. *deprecated* use `tls` variants
+ * @param {boolean} [options.tls=false] Enable TLS connections
+ * @param {boolean} [options.tlsInsecure=false] Relax TLS constraints, disabling validation
+ * @param {string} [options.tlsCAFile] A path to file with either a single or bundle of certificate authorities to be considered trusted when making a TLS connection
+ * @param {string} [options.tlsCertificateKeyFile] A path to the client certificate file or the client private key file; in the case that they both are needed, the files should be concatenated
+ * @param {string} [options.tlsCertificateKeyFilePassword] The password to decrypt the client private key to be used for TLS connections
+ * @param {boolean} [options.tlsAllowInvalidCertificates] Specifies whether or not the driver should error when the server’s TLS certificate is invalid
+ * @param {boolean} [options.tlsAllowInvalidHostnames] Specifies whether or not the driver should error when there is a mismatch between the server’s hostname and the hostname specified by the TLS certificate
  * @param {boolean} [options.autoReconnect=true] Enable autoReconnect for single server instances
  * @param {boolean} [options.noDelay=true] TCP Connection no delay
  * @param {boolean} [options.keepAlive=true] TCP Connection keep alive enabled
  * @param {number} [options.keepAliveInitialDelay=30000] The number of milliseconds to wait before initiating keepAlive on the TCP socket
- * @param {number} [options.connectTimeoutMS=30000] TCP Connection timeout setting
+ * @param {number} [options.connectTimeoutMS=10000] How long to wait for a connection to be established before timing out
+ * @param {number} [options.socketTimeoutMS=360000] How long a send or receive on a socket can take before timing out
  * @param {number} [options.family] Version of IP stack. Can be 4, 6 or null (default).
  * If null, will attempt to connect with IPv6, and will fall back to IPv4 on failure
- * @param {number} [options.socketTimeoutMS=360000] TCP Socket timeout setting
  * @param {number} [options.reconnectTries=30] Server attempt to reconnect #times
  * @param {number} [options.reconnectInterval=1000] Server will wait # milliseconds between retries
  * @param {boolean} [options.ha=true] Control if high availability monitoring runs for Replicaset or Mongos proxies
@@ -127,7 +129,6 @@ const CloseOperation = require('./operations/close');
  * @param {boolean} [options.promoteBuffers=false] Promotes Binary BSON values to native Node Buffers
  * @param {boolean} [options.promoteLongs=true] Promotes long values to number if they fit inside the 53 bits resolution
  * @param {boolean} [options.domainsEnabled=false] Enable the wrapping of the callback in the current domain, disabled by default to avoid perf hit
- * @param {boolean|function} [options.checkServerIdentity=true] Ensure we check server identify during SSL, set to false to disable checking. Only works for Node 0.12.x or higher. You can pass in a boolean or your own checkServerIdentity override function
  * @param {object} [options.validateOptions=false] Validate MongoClient passed in options for correctness
  * @param {string} [options.appname=undefined] The name of the application that created this MongoClient instance. MongoDB 3.4 and newer will print this value in the server log upon establishing each connection. It is also recorded in the slow query log and profile collections
  * @param {string} [options.auth.user=undefined] The username for auth
@@ -140,9 +141,10 @@ const CloseOperation = require('./operations/close');
  * @param {boolean} [options.auto_reconnect=true] Enable auto reconnecting for single server instances
  * @param {boolean} [options.monitorCommands=false] Enable command monitoring for this client
  * @param {number} [options.minSize] If present, the connection pool will be initialized with minSize connections, and will never dip below minSize connections
- * @param {boolean} [options.useNewUrlParser=false] Determines whether or not to use the new url parser. Enables the new, spec-compliant, url parser shipped in the core driver. This url parser fixes a number of problems with the original parser, and aims to outright replace that parser in the near future.
+ * @param {boolean} [options.useNewUrlParser=true] Determines whether or not to use the new url parser. Enables the new, spec-compliant, url parser shipped in the core driver. This url parser fixes a number of problems with the original parser, and aims to outright replace that parser in the near future. Defaults to true, and must be explicitly set to false to use the legacy url parser.
  * @param {boolean} [options.useUnifiedTopology] Enables the new unified topology layer
- * @param {AutoEncryptionOptions} [options.autoEncryption] Optionally enable client side auto encryption
+ * @param {AutoEncrypter~AutoEncryptionOptions} [options.autoEncryption] Optionally enable client side auto encryption
+ * @param {DriverInfoOptions} [options.driverInfo] Allows a wrapping driver to amend the client metadata generated by the driver to include information about the wrapping driver
  * @param {MongoClient~connectCallback} [callback] The command result callback
  * @return {MongoClient} a MongoClient instance
  */
@@ -155,18 +157,12 @@ function MongoClient(url, options) {
   this.s = {
     url: url,
     options: options || {},
-    promiseLibrary: null,
+    promiseLibrary: (options && options.promiseLibrary) || Promise,
     dbCache: new Map(),
-    sessions: [],
+    sessions: new Set(),
     writeConcern: WriteConcern.fromOptions(options),
     namespace: new MongoDBNamespace('admin')
   };
-
-  // Get the promiseLibrary
-  const promiseLibrary = this.s.options.promiseLibrary || Promise;
-
-  // Add the promise to the internal state
-  this.s.promiseLibrary = promiseLibrary;
 }
 
 /**
@@ -211,9 +207,16 @@ MongoClient.prototype.connect = function(callback) {
     throw new TypeError('`connect` only accepts a callback');
   }
 
-  const operation = new ConnectOperation(this);
+  const client = this;
+  return maybePromise(this, callback, cb => {
+    const err = validOptions(client.s.options);
+    if (err) return cb(err);
 
-  return executeOperation(this, operation, callback);
+    connect(client, client.s.url, client.s.options, err => {
+      if (err) return cb(err);
+      cb(null, client);
+    });
+  });
 };
 
 MongoClient.prototype.logout = deprecate(function(options, callback) {
@@ -229,9 +232,41 @@ MongoClient.prototype.logout = deprecate(function(options, callback) {
  * @return {Promise} returns Promise if no callback passed
  */
 MongoClient.prototype.close = function(force, callback) {
-  if (typeof force === 'function') (callback = force), (force = false);
-  const operation = new CloseOperation(this, force);
-  return executeOperation(this, operation, callback);
+  if (typeof force === 'function') {
+    callback = force;
+    force = false;
+  }
+
+  const client = this;
+  return maybePromise(this, callback, cb => {
+    const completeClose = err => {
+      client.emit('close', client);
+
+      if (!(client.topology instanceof NativeTopology)) {
+        for (const item of client.s.dbCache) {
+          item[1].emit('close', client);
+        }
+      }
+
+      client.removeAllListeners('close');
+      cb(err);
+    };
+
+    if (client.topology == null) {
+      completeClose();
+      return;
+    }
+
+    client.topology.close(force, err => {
+      const autoEncrypter = client.topology.s.options.autoEncrypter;
+      if (!autoEncrypter) {
+        completeClose(err);
+        return;
+      }
+
+      autoEncrypter.teardown(force, err2 => completeClose(err || err2));
+    });
+  });
 };
 
 /**
@@ -308,21 +343,29 @@ MongoClient.prototype.isConnected = function(options) {
  * @param {string} url The connection URI string
  * @param {object} [options] Optional settings
  * @param {number} [options.poolSize=5] The maximum size of the individual server pool
- * @param {boolean} [options.ssl=false] Enable SSL connection.
- * @param {boolean} [options.sslValidate=false] Validate mongod server certificate against Certificate Authority
- * @param {buffer} [options.sslCA=undefined] SSL Certificate store binary buffer
- * @param {buffer} [options.sslCert=undefined] SSL Certificate binary buffer
- * @param {buffer} [options.sslKey=undefined] SSL Key file binary buffer
- * @param {string} [options.sslPass=undefined] SSL Certificate pass phrase
- * @param {buffer} [options.sslCRL=undefined] SSL Certificate revocation list binary buffer
+ * @param {boolean} [options.ssl=false] Enable SSL connection. *deprecated* use `tls` variants
+ * @param {boolean} [options.sslValidate=false] Validate mongod server certificate against Certificate Authority *deprecated* use `tls` variants
+ * @param {buffer} [options.sslCA=undefined] SSL Certificate store binary buffer *deprecated* use `tls` variants
+ * @param {buffer} [options.sslCert=undefined] SSL Certificate binary buffer *deprecated* use `tls` variants
+ * @param {buffer} [options.sslKey=undefined] SSL Key file binary buffer *deprecated* use `tls` variants
+ * @param {string} [options.sslPass=undefined] SSL Certificate pass phrase *deprecated* use `tls` variants
+ * @param {buffer} [options.sslCRL=undefined] SSL Certificate revocation list binary buffer *deprecated* use `tls` variants
+ * @param {boolean|function} [options.checkServerIdentity=true] Ensure we check server identify during SSL, set to false to disable checking. Only works for Node 0.12.x or higher. You can pass in a boolean or your own checkServerIdentity override function. *deprecated* use `tls` variants
+ * @param {boolean} [options.tls=false] Enable TLS connections
+ * @param {boolean} [options.tlsInsecure=false] Relax TLS constraints, disabling validation
+ * @param {string} [options.tlsCAFile] A path to file with either a single or bundle of certificate authorities to be considered trusted when making a TLS connection
+ * @param {string} [options.tlsCertificateKeyFile] A path to the client certificate file or the client private key file; in the case that they both are needed, the files should be concatenated
+ * @param {string} [options.tlsCertificateKeyFilePassword] The password to decrypt the client private key to be used for TLS connections
+ * @param {boolean} [options.tlsAllowInvalidCertificates] Specifies whether or not the driver should error when the server’s TLS certificate is invalid
+ * @param {boolean} [options.tlsAllowInvalidHostnames] Specifies whether or not the driver should error when there is a mismatch between the server’s hostname and the hostname specified by the TLS certificate
  * @param {boolean} [options.autoReconnect=true] Enable autoReconnect for single server instances
  * @param {boolean} [options.noDelay=true] TCP Connection no delay
  * @param {boolean} [options.keepAlive=true] TCP Connection keep alive enabled
  * @param {boolean} [options.keepAliveInitialDelay=30000] The number of milliseconds to wait before initiating keepAlive on the TCP socket
- * @param {number} [options.connectTimeoutMS=30000] TCP Connection timeout setting
+ * @param {number} [options.connectTimeoutMS=10000] How long to wait for a connection to be established before timing out
+ * @param {number} [options.socketTimeoutMS=360000] How long a send or receive on a socket can take before timing out
  * @param {number} [options.family] Version of IP stack. Can be 4, 6 or null (default).
  * If null, will attempt to connect with IPv6, and will fall back to IPv4 on failure
- * @param {number} [options.socketTimeoutMS=360000] TCP Socket timeout setting
  * @param {number} [options.reconnectTries=30] Server attempt to reconnect #times
  * @param {number} [options.reconnectInterval=1000] Server will wait # milliseconds between retries
  * @param {boolean} [options.ha=true] Control if high availability monitoring runs for Replicaset or Mongos proxies
@@ -352,7 +395,6 @@ MongoClient.prototype.isConnected = function(options) {
  * @param {boolean} [options.promoteBuffers=false] Promotes Binary BSON values to native Node Buffers
  * @param {boolean} [options.promoteLongs=true] Promotes long values to number if they fit inside the 53 bits resolution
  * @param {boolean} [options.domainsEnabled=false] Enable the wrapping of the callback in the current domain, disabled by default to avoid perf hit
- * @param {boolean|function} [options.checkServerIdentity=true] Ensure we check server identify during SSL, set to false to disable checking. Only works for Node 0.12.x or higher. You can pass in a boolean or your own checkServerIdentity override function
  * @param {object} [options.validateOptions=false] Validate MongoClient passed in options for correctness
  * @param {string} [options.appname=undefined] The name of the application that created this MongoClient instance. MongoDB 3.4 and newer will print this value in the server log upon establishing each connection. It is also recorded in the slow query log and profile collections
  * @param {string} [options.auth.user=undefined] The username for auth
@@ -446,7 +488,7 @@ MongoClient.prototype.withSession = function(options, operation) {
  * @param {string} [options.fullDocument='default'] Allowed values: ‘default’, ‘updateLookup’. When set to ‘updateLookup’, the change stream will include both a delta describing the changes to the document, as well as a copy of the entire document that was changed from some time after the change occurred.
  * @param {object} [options.resumeAfter] Specifies the logical starting point for the new change stream. This should be the _id field from a previously returned change stream document.
  * @param {number} [options.maxAwaitTimeMS] The maximum amount of time for the server to wait on new documents to satisfy a change stream query
- * @param {number} [options.batchSize] The number of documents to return per batch. See {@link https://docs.mongodb.com/manual/reference/command/aggregate|aggregation documentation}.
+ * @param {number} [options.batchSize=1000] The number of documents to return per batch. See {@link https://docs.mongodb.com/manual/reference/command/aggregate|aggregation documentation}.
  * @param {object} [options.collation] Specify collation settings for operation. See {@link https://docs.mongodb.com/manual/reference/command/aggregate|aggregation documentation}.
  * @param {ReadPreference} [options.readPreference] The read preference. See {@link https://docs.mongodb.com/manual/reference/read-preference|read preference documentation}.
  * @param {Timestamp} [options.startAtOperationTime] receive change events that occur after the specified timestamp
