@@ -1,8 +1,5 @@
 'use strict';
 
-const OperationBase = require('./operation').OperationBase;
-const defineAspects = require('./operation').defineAspects;
-const Aspect = require('./operation').Aspect;
 const deprecate = require('util').deprecate;
 const Logger = require('../core').Logger;
 const MongoCredentials = require('../core').MongoCredentials;
@@ -15,6 +12,10 @@ const ReadPreference = require('../core').ReadPreference;
 const ReplSet = require('../topologies/replset');
 const Server = require('../topologies/server');
 const ServerSessionPool = require('../core').Sessions.ServerSessionPool;
+const emitDeprecationWarning = require('../utils').emitDeprecationWarning;
+const fs = require('fs');
+const BSON = require('../core/connection/utils').retrieveBSON();
+const CMAP_EVENT_NAMES = require('../cmap/events').CMAP_EVENT_NAMES;
 
 let client;
 function loadClient() {
@@ -137,7 +138,20 @@ const validOptionNames = [
   'useUnifiedTopology',
   'serverSelectionTimeoutMS',
   'useRecoveryToken',
-  'autoEncryption'
+  'autoEncryption',
+  'driverInfo',
+  'tls',
+  'tlsInsecure',
+  'tlsinsecure',
+  'tlsAllowInvalidCertificates',
+  'tlsAllowInvalidHostnames',
+  'tlsCAFile',
+  'tlsCertificateFile',
+  'tlsCertificateKeyFile',
+  'tlsCertificateKeyFilePassword',
+  'minHeartbeatFrequencyMS',
+  'heartbeatFrequencyMS',
+  'waitQueueTimeoutMS'
 ];
 
 const ignoreOptionNames = ['native_parser'];
@@ -174,28 +188,6 @@ const LEGACY_OPTIONS_MAP = validOptionNames.reduce((obj, name) => {
   return obj;
 }, {});
 
-class ConnectOperation extends OperationBase {
-  constructor(mongoClient) {
-    super();
-
-    this.mongoClient = mongoClient;
-  }
-
-  execute(callback) {
-    const mongoClient = this.mongoClient;
-    const err = validOptions(mongoClient.s.options);
-
-    // Did we have a validation error
-    if (err) return callback(err);
-    // Fallback to callback based connect
-    connect(mongoClient, mongoClient.s.url, mongoClient.s.options, err => {
-      if (err) return callback(err);
-      callback(null, mongoClient);
-    });
-  }
-}
-defineAspects(ConnectOperation, [Aspect.SKIP_SESSION]);
-
 function addListeners(mongoClient, topology) {
   topology.on('authenticated', createListener(mongoClient, 'authenticated'));
   topology.on('error', createListener(mongoClient, 'error'));
@@ -210,10 +202,10 @@ function addListeners(mongoClient, topology) {
 
 function assignTopology(client, topology) {
   client.topology = topology;
-  topology.s.sessionPool =
-    topology instanceof NativeTopology
-      ? new ServerSessionPool(topology)
-      : new ServerSessionPool(topology.s.coreTopology);
+
+  if (!(topology instanceof NativeTopology)) {
+    topology.s.sessionPool = new ServerSessionPool(topology.s.coreTopology);
+  }
 }
 
 // Clear out all events
@@ -241,8 +233,20 @@ function collectEvents(mongoClient, topology) {
   return collectedEvents;
 }
 
+function resolveTLSOptions(options) {
+  if (options.tls == null) {
+    return;
+  }
+
+  ['sslCA', 'sslKey', 'sslCert'].forEach(optionName => {
+    if (options[optionName]) {
+      options[optionName] = fs.readFileSync(options[optionName]);
+    }
+  });
+}
+
 const emitDeprecationForNonUnifiedTopology = deprecate(() => {},
-'current Server Discovery and Monitoring engine is deprecated, and will be removed in a future version. ' + 'To use the new Server Discover and Monitoring engine, pass option { useUnifiedTopology: true } to MongoClient.connect.');
+'current Server Discovery and Monitoring engine is deprecated, and will be removed in a future version. ' + 'To use the new Server Discover and Monitoring engine, pass option { useUnifiedTopology: true } to the MongoClient constructor.');
 
 function connect(mongoClient, url, options, callback) {
   options = Object.assign({}, options);
@@ -260,8 +264,10 @@ function connect(mongoClient, url, options, callback) {
     return connectWithUrl(mongoClient, url, options, connectCallback);
   }
 
-  const parseFn = options.useNewUrlParser ? parse : legacyParse;
-  const transform = options.useNewUrlParser ? transformUrlOptions : legacyTransformUrlOptions;
+  const useNewUrlParser = options.useNewUrlParser !== false;
+
+  const parseFn = useNewUrlParser ? parse : legacyParse;
+  const transform = useNewUrlParser ? transformUrlOptions : legacyTransformUrlOptions;
 
   parseFn(url, options, (err, _object) => {
     // Do not attempt to connect if parsing error
@@ -275,14 +281,18 @@ function connect(mongoClient, url, options, callback) {
 
     // Check if we have connection and socket timeout set
     if (_finalOptions.socketTimeoutMS == null) _finalOptions.socketTimeoutMS = 360000;
-    if (_finalOptions.connectTimeoutMS == null) _finalOptions.connectTimeoutMS = 30000;
+    if (_finalOptions.connectTimeoutMS == null) _finalOptions.connectTimeoutMS = 10000;
     if (_finalOptions.retryWrites == null) _finalOptions.retryWrites = true;
     if (_finalOptions.useRecoveryToken == null) _finalOptions.useRecoveryToken = true;
+    if (_finalOptions.readPreference == null) _finalOptions.readPreference = 'primary';
 
     if (_finalOptions.db_options && _finalOptions.db_options.auth) {
       delete _finalOptions.db_options.auth;
     }
 
+    // resolve tls options if needed
+    resolveTLSOptions(_finalOptions);
+
     // Store the merged options object
     mongoClient.s.options = _finalOptions;
 
@@ -434,6 +444,30 @@ function createServer(mongoClient, options, callback) {
   });
 }
 
+const DEPRECATED_UNIFIED_EVENTS = new Set([
+  'reconnect',
+  'reconnectFailed',
+  'attemptReconnect',
+  'joined',
+  'left',
+  'ping',
+  'ha',
+  'all',
+  'fullsetup',
+  'open'
+]);
+
+function registerDeprecatedEventNotifiers(client) {
+  client.on('newListener', eventName => {
+    if (DEPRECATED_UNIFIED_EVENTS.has(eventName)) {
+      emitDeprecationWarning(
+        `The \`${eventName}\` event is no longer supported by the unified topology, please read more by visiting http://bit.ly/2D8WfT6`,
+        'DeprecationWarning'
+      );
+    }
+  });
+}
+
 function createTopology(mongoClient, topologyType, options, callback) {
   // Pass in the promise library
   options.promiseLibrary = mongoClient.s.promiseLibrary;
@@ -444,36 +478,8 @@ function createTopology(mongoClient, topologyType, options, callback) {
   // Set default options
   const servers = translateOptions(options, translationOptions);
 
-  // Create the topology
-  let topology;
-  if (topologyType === 'mongos') {
-    topology = new Mongos(servers, options);
-  } else if (topologyType === 'replicaset') {
-    topology = new ReplSet(servers, options);
-  } else if (topologyType === 'unified') {
-    topology = new NativeTopology(options.servers, options);
-  }
-
-  // Add listeners
-  addListeners(mongoClient, topology);
-
-  // Propagate the events to the client
-  relayEvents(mongoClient, topology);
-
-  // Open the connection
-  topology.connect(options, (err, newTopology) => {
-    if (err) {
-      topology.close(true);
-      return callback(err);
-    }
-
-    assignTopology(mongoClient, newTopology);
-    if (options.autoEncryption == null) {
-      callback(null, newTopology);
-      return;
-    }
-
-    // setup for client side encryption
+  // determine CSFLE support
+  if (options.autoEncryption != null) {
     let AutoEncrypter;
     try {
       require.resolve('mongodb-client-encryption');
@@ -485,19 +491,100 @@ function createTopology(mongoClient, topologyType, options, callback) {
       );
       return;
     }
+
     try {
-      AutoEncrypter = require('mongodb-client-encryption')(require('../../index')).AutoEncrypter;
+      let mongodbClientEncryption = require('mongodb-client-encryption');
+      if (typeof mongodbClientEncryption.extension !== 'function') {
+        callback(
+          new MongoError(
+            'loaded version of `mongodb-client-encryption` does not have property `extension`. Please make sure you are loading the correct version of `mongodb-client-encryption`'
+          )
+        );
+      }
+      AutoEncrypter = mongodbClientEncryption.extension(require('../../index')).AutoEncrypter;
     } catch (err) {
       callback(err);
       return;
     }
 
-    const mongoCryptOptions = Object.assign({}, options.autoEncryption);
-    topology.s.options.autoEncrypter = new AutoEncrypter(mongoClient, mongoCryptOptions);
-    topology.s.options.autoEncrypter.init(err => {
-      if (err) return callback(err, null);
-      callback(null, newTopology);
+    const mongoCryptOptions = Object.assign(
+      {
+        bson:
+          options.bson ||
+          new BSON([
+            BSON.Binary,
+            BSON.Code,
+            BSON.DBRef,
+            BSON.Decimal128,
+            BSON.Double,
+            BSON.Int32,
+            BSON.Long,
+            BSON.Map,
+            BSON.MaxKey,
+            BSON.MinKey,
+            BSON.ObjectId,
+            BSON.BSONRegExp,
+            BSON.Symbol,
+            BSON.Timestamp
+          ])
+      },
+      options.autoEncryption
+    );
+
+    options.autoEncrypter = new AutoEncrypter(mongoClient, mongoCryptOptions);
+  }
+
+  // Create the topology
+  let topology;
+  if (topologyType === 'mongos') {
+    topology = new Mongos(servers, options);
+  } else if (topologyType === 'replicaset') {
+    topology = new ReplSet(servers, options);
+  } else if (topologyType === 'unified') {
+    topology = new NativeTopology(options.servers, options);
+    registerDeprecatedEventNotifiers(mongoClient);
+  }
+
+  // Add listeners
+  addListeners(mongoClient, topology);
+
+  // Propagate the events to the client
+  relayEvents(mongoClient, topology);
+
+  // Open the connection
+  assignTopology(mongoClient, topology);
+
+  // initialize CSFLE if requested
+  if (options.autoEncrypter) {
+    options.autoEncrypter.init(err => {
+      if (err) {
+        callback(err);
+        return;
+      }
+
+      topology.connect(options, err => {
+        if (err) {
+          topology.close(true);
+          callback(err);
+          return;
+        }
+
+        callback(undefined, topology);
+      });
     });
+
+    return;
+  }
+
+  // otherwise connect normally
+  topology.connect(options, err => {
+    if (err) {
+      topology.close(true);
+      return callback(err);
+    }
+
+    callback(undefined, topology);
+    return;
   });
 }
 
@@ -512,7 +599,7 @@ function createUnifiedOptions(finalOptions, options) {
     'rs_options',
     'mongos_options'
   ];
-  const noMerge = ['readconcern', 'compression'];
+  const noMerge = ['readconcern', 'compression', 'autoencryption'];
 
   for (const name in options) {
     if (noMerge.indexOf(name.toLowerCase()) !== -1) {
@@ -590,23 +677,28 @@ function mergeOptions(target, source, flatten) {
 
 function relayEvents(mongoClient, topology) {
   const serverOrCommandEvents = [
+    // APM
+    'commandStarted',
+    'commandSucceeded',
+    'commandFailed',
+
+    // SDAM
     'serverOpening',
+    'serverClosed',
     'serverDescriptionChanged',
     'serverHeartbeatStarted',
     'serverHeartbeatSucceeded',
     'serverHeartbeatFailed',
-    'serverClosed',
     'topologyOpening',
     'topologyClosed',
     'topologyDescriptionChanged',
-    'commandStarted',
-    'commandSucceeded',
-    'commandFailed',
+
+    // Legacy
     'joined',
     'left',
     'ping',
     'ha'
-  ];
+  ].concat(CMAP_EVENT_NAMES);
 
   serverOrCommandEvents.forEach(event => {
     topology.on(event, (object1, object2) => {
@@ -650,16 +742,16 @@ function transformUrlOptions(_object) {
     object.dbName = _object.defaultDatabase;
   }
 
-  if (object.maxpoolsize) {
-    object.poolSize = object.maxpoolsize;
+  if (object.maxPoolSize) {
+    object.poolSize = object.maxPoolSize;
   }
 
-  if (object.readconcernlevel) {
-    object.readConcern = new ReadConcern(object.readconcernlevel);
+  if (object.readConcernLevel) {
+    object.readConcern = new ReadConcern(object.readConcernLevel);
   }
 
-  if (object.wtimeoutms) {
-    object.wtimeout = object.wtimeoutms;
+  if (object.wTimeoutMS) {
+    object.wtimeout = object.wTimeoutMS;
   }
 
   if (_object.srvHost) {
@@ -689,7 +781,7 @@ function translateOptions(options, translationOptions) {
 
   // Set the socket and connection timeouts
   if (options.socketTimeoutMS == null) options.socketTimeoutMS = 360000;
-  if (options.connectTimeoutMS == null) options.connectTimeoutMS = 30000;
+  if (options.connectTimeoutMS == null) options.connectTimeoutMS = 10000;
 
   if (!translationOptions.createServers) {
     return;
@@ -703,4 +795,4 @@ function translateOptions(options, translationOptions) {
   });
 }
 
-module.exports = ConnectOperation;
+module.exports = { validOptions, connect };
