@@ -17,16 +17,20 @@ class FindOneOperation extends OperationBase {
     const query = this.query;
     const options = this.options;
 
-    const cursor = coll
-      .find(query, options)
-      .limit(-1)
-      .batchSize(1);
+    try {
+      const cursor = coll
+        .find(query, options)
+        .limit(-1)
+        .batchSize(1);
 
-    // Return the item
-    cursor.next((err, item) => {
-      if (err != null) return handleCallback(callback, toError(err), null);
-      handleCallback(callback, null, item);
-    });
+      // Return the item
+      cursor.next((err, item) => {
+        if (err != null) return handleCallback(callback, toError(err), null);
+        handleCallback(callback, null, item);
+      });
+    } catch (e) {
+      callback(e);
+    }
   }
 }
 
