@@ -2,7 +2,6 @@
 
 const Aspect = {
   READ_OPERATION: Symbol('READ_OPERATION'),
-  SKIP_SESSION: Symbol('SKIP_SESSION'),
   WRITE_OPERATION: Symbol('WRITE_OPERATION'),
   RETRYABLE: Symbol('RETRYABLE'),
   EXECUTE_WITH_SELECTION: Symbol('EXECUTE_WITH_SELECTION')
@@ -12,8 +11,7 @@ const Aspect = {
  * This class acts as a parent class for any operation and is responsible for setting this.options,
  * as well as setting and getting a session.
  * Additionally, this class implements `hasAspect`, which determines whether an operation has
- * a specific aspect, including `SKIP_SESSION` and other aspects to encode retryability
- * and other functionality.
+ * a specific aspect.
  */
 class OperationBase {
   constructor(options) {
