@@ -1,5 +1,4 @@
 'use strict';
-
 const MongoError = require('./core/error').MongoError;
 const ReadPreference = require('./core/topologies/read_preference');
 const WriteConcern = require('./write_concern');
@@ -685,6 +684,56 @@ class MongoDBNamespace {
   }
 }
 
+function* makeCounter(seed) {
+  let count = seed || 0;
+  while (true) {
+    const newCount = count;
+    count += 1;
+    yield newCount;
+  }
+}
+
+/**
+ * Helper function for either accepting a callback, or returning a promise
+ *
+ * @param {Object} parent an instance of parent with promiseLibrary.
+ * @param {object} parent.s an object containing promiseLibrary.
+ * @param {function} parent.s.promiseLibrary an object containing promiseLibrary.
+ * @param {[Function]} callback an optional callback.
+ * @param {Function} fn A function that takes a callback
+ * @returns {Promise|void} Returns nothing if a callback is supplied, else returns a Promise.
+ */
+function maybePromise(parent, callback, fn) {
+  const PromiseLibrary = (parent && parent.s && parent.s.promiseLibrary) || Promise;
+
+  let result;
+  if (typeof callback !== 'function') {
+    result = new PromiseLibrary((resolve, reject) => {
+      callback = (err, res) => {
+        if (err) return reject(err);
+        resolve(res);
+      };
+    });
+  }
+
+  fn(function(err, res) {
+    if (err != null) {
+      try {
+        callback(err);
+      } catch (error) {
+        return process.nextTick(() => {
+          throw error;
+        });
+      }
+      return;
+    }
+
+    callback(err, res);
+  });
+
+  return result;
+}
+
 module.exports = {
   filterOptions,
   mergeOptions,
@@ -712,5 +761,8 @@ module.exports = {
   deprecateOptions,
   SUPPORTS,
   MongoDBNamespace,
-  resolveReadPreference
+  resolveReadPreference,
+  emitDeprecationWarning,
+  makeCounter,
+  maybePromise
 };
