@@ -1,5 +1,5 @@
 'use strict';
-const Promise = require('bluebird');
+
 const path = require('path');
 const fs = require('fs');
 const chai = require('chai');
@@ -14,6 +14,17 @@ chai.config.includeStack = true;
 chai.config.showDiff = true;
 chai.config.truncateThreshold = 0;
 
+// Promise.try alternative https://stackoverflow.com/questions/60624081/promise-try-without-bluebird/60624164?noredirect=1#comment107255389_60624164
+function promiseTry(callback) {
+  return new Promise((resolve, reject) => {
+    try {
+      resolve(callback());
+    } catch (e) {
+      reject(e);
+    }
+  });
+}
+
 function escape(string) {
   return string.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');
 }
@@ -233,6 +244,7 @@ function runTestSuiteTest(configuration, spec, context) {
   // test-specific client options
   clientOptions.autoReconnect = false;
   clientOptions.haInterval = 100;
+  clientOptions.minHeartbeatFrequencyMS = 100;
   clientOptions.useRecoveryToken = true;
 
   // TODO: this should be configured by `newClient` and env variables
@@ -570,7 +582,7 @@ function testOperation(operation, obj, context, options) {
     opPromise = cursor.toArray();
   } else {
     // wrap this in a `Promise.try` because some operations might throw
-    opPromise = Promise.try(() => obj[operationName].apply(obj, args));
+    opPromise = promiseTry(() => obj[operationName].apply(obj, args));
   }
 
   if (operation.error) {
